# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base form classes
"""

import logging
from collections import OrderedDict

import sqlalchemy as sa
from sqlalchemy import orm

import colander
import deform
from colanderalchemy import SQLAlchemySchemaNode
from pyramid.renderers import render
from webhelpers2.html import HTML

from wuttaweb.util import FieldList, get_form_data, get_model_fields, make_json_safe


log = logging.getLogger(__name__)


class Form:
    """
    Base class for all forms.

    :param request: Reference to current :term:`request` object.

    :param fields: List of field names for the form.  This is
       optional; if not specified an attempt will be made to deduce
       the list automatically.  See also :attr:`fields`.

    :param schema: Colander-based schema object for the form.  This is
       optional; if not specified an attempt will be made to construct
       one automatically.  See also :meth:`get_schema()`.

    :param labels: Optional dict of default field labels.

    .. note::

       Some parameters are not explicitly described above.  However
       their corresponding attributes are described below.

    Form instances contain the following attributes:

    .. attribute:: request

       Reference to current :term:`request` object.

    .. attribute:: fields

       :class:`~wuttaweb.util.FieldList` instance containing string
       field names for the form.  By default, fields will appear in
       the same order as they are in this list.

       See also :meth:`set_fields()`.

    .. attribute:: schema

       :class:`colander:colander.Schema` object for the form.  This is
       optional; if not specified an attempt will be made to construct
       one automatically.

       See also :meth:`get_schema()`.

    .. attribute:: model_class

       Model class for the form, if applicable.  When set, this is
       usually a SQLAlchemy mapped class.  This (or
       :attr:`model_instance`) may be used instead of specifying the
       :attr:`schema`.

    .. attribute:: model_instance

       Optional instance from which initial form data should be
       obtained.  In simple cases this might be a dict, or maybe an
       instance of :attr:`model_class`.

       Note that this also may be used instead of specifying the
       :attr:`schema`, if the instance belongs to a class which is
       SQLAlchemy-mapped.  (In that case :attr:`model_class` can be
       determined automatically.)

    .. attribute:: nodes

       Dict of node overrides, used to construct the form in
       :meth:`get_schema()`.

       See also :meth:`set_node()`.

    .. attribute:: widgets

       Dict of widget overrides, used to construct the form in
       :meth:`get_schema()`.

       See also :meth:`set_widget()`.

    .. attribute:: validators

       Dict of node validators, used to construct the form in
       :meth:`get_schema()`.

       See also :meth:`set_validator()`.

    .. attribute:: defaults

       Dict of default field values, used to construct the form in
       :meth:`get_schema()`.

       See also :meth:`set_default()`.

    .. attribute:: readonly

       Boolean indicating the form does not allow submit.  In practice
       this means there will not even be a ``<form>`` tag involved.

       Default for this is ``False`` in which case the ``<form>`` tag
       will exist and submit is allowed.

    .. attribute:: readonly_fields

       A :class:`~python:set` of field names which should be readonly.
       Each will still be rendered but with static value text and no
       widget.

       This is only applicable if :attr:`readonly` is ``False``.

       See also :meth:`set_readonly()` and :meth:`is_readonly()`.

    .. attribute:: required_fields

       A dict of "required" field flags.  Keys are field names, and
       values are boolean flags indicating whether the field is
       required.

       Depending on :attr:`schema`, some fields may be "(not)
       required" by default.  However ``required_fields`` keeps track
       of any "overrides" per field.

       See also :meth:`set_required()` and :meth:`is_required()`.

    .. attribute:: action_method

       HTTP method to use when submitting form; ``'post'`` is default.

    .. attribute:: action_url

       String URL to which the form should be submitted, if applicable.

    .. attribute:: reset_url

       String URL to which the reset button should "always" redirect,
       if applicable.

       This is null by default, in which case it will use standard
       browser behavior for the form reset button (if shown).  See
       also :attr:`show_button_reset`.

    .. attribute:: cancel_url

       String URL to which the Cancel button should "always" redirect,
       if applicable.

       Code should not access this directly, but instead call
       :meth:`get_cancel_url()`.

    .. attribute:: cancel_url_fallback

       String URL to which the Cancel button should redirect, if
       referrer cannot be determined from request.

       Code should not access this directly, but instead call
       :meth:`get_cancel_url()`.

    .. attribute:: vue_tagname

       String name for Vue component tag.  By default this is
       ``'wutta-form'``.  See also :meth:`render_vue_tag()`.

       See also :attr:`vue_component`.

    .. attribute:: align_buttons_right

       Flag indicating whether the buttons (submit, cancel etc.)
       should be aligned to the right of the area below the form.  If
       not set, the buttons are left-aligned.

    .. attribute:: auto_disable_submit

       Flag indicating whether the submit button should be
       auto-disabled, whenever the form is submitted.

    .. attribute:: button_label_submit

       String label for the form submit button.  Default is ``"Save"``.

    .. attribute:: button_icon_submit

       String icon name for the form submit button.  Default is ``'save'``.

    .. attribute:: button_type_submit

       Buefy type for the submit button.  Default is ``'is-primary'``,
       so for example:

       .. code-block:: html

          <b-button type="is-primary"
                    native-type="submit">
            Save
          </b-button>

       See also the `Buefy docs
       <https://buefy.org/documentation/button/#api-view>`_.

    .. attribute:: show_button_reset

       Flag indicating whether a Reset button should be shown.
       Default is ``False``.

       Unless there is a :attr:`reset_url`, the reset button will use
       standard behavior per the browser.

    .. attribute:: show_button_cancel

       Flag indicating whether a Cancel button should be shown.
       Default is ``True``.

    .. attribute:: button_label_cancel

       String label for the form cancel button.  Default is
       ``"Cancel"``.

    .. attribute:: auto_disable_cancel

       Flag indicating whether the cancel button should be
       auto-disabled, whenever the button is clicked.  Default is
       ``True``.

    .. attribute:: validated

       If the :meth:`validate()` method was called, and it succeeded,
       this will be set to the validated data dict.

       Note that in all other cases, this attribute may not exist.
    """

    def __init__(
            self,
            request,
            fields=None,
            schema=None,
            model_class=None,
            model_instance=None,
            nodes={},
            widgets={},
            validators={},
            defaults={},
            readonly=False,
            readonly_fields=[],
            required_fields={},
            labels={},
            action_method='post',
            action_url=None,
            reset_url=None,
            cancel_url=None,
            cancel_url_fallback=None,
            vue_tagname='wutta-form',
            align_buttons_right=False,
            auto_disable_submit=True,
            button_label_submit="Save",
            button_icon_submit='save',
            button_type_submit='is-primary',
            show_button_reset=False,
            show_button_cancel=True,
            button_label_cancel="Cancel",
            auto_disable_cancel=True,
    ):
        self.request = request
        self.schema = schema
        self.nodes = nodes or {}
        self.widgets = widgets or {}
        self.validators = validators or {}
        self.defaults = defaults or {}
        self.readonly = readonly
        self.readonly_fields = set(readonly_fields or [])
        self.required_fields = required_fields or {}
        self.labels = labels or {}
        self.action_method = action_method
        self.action_url = action_url
        self.cancel_url = cancel_url
        self.cancel_url_fallback = cancel_url_fallback
        self.reset_url = reset_url
        self.vue_tagname = vue_tagname
        self.align_buttons_right = align_buttons_right
        self.auto_disable_submit = auto_disable_submit
        self.button_label_submit = button_label_submit
        self.button_icon_submit = button_icon_submit
        self.button_type_submit = button_type_submit
        self.show_button_reset = show_button_reset
        self.show_button_cancel = show_button_cancel
        self.button_label_cancel = button_label_cancel
        self.auto_disable_cancel = auto_disable_cancel

        self.config = self.request.wutta_config
        self.app = self.config.get_app()

        self.model_class = model_class
        self.model_instance = model_instance
        if self.model_instance and not self.model_class:
            if type(self.model_instance) is not dict:
                self.model_class = type(self.model_instance)

        self.set_fields(fields or self.get_fields())
        self.set_default_widgets()

        # nb. this tracks grid JSON data for inclusion in page template
        self.grid_vue_context = OrderedDict()

    def __contains__(self, name):
        """
        Custom logic for the ``in`` operator, to allow easily checking
        if the form contains a given field::

           myform = Form()
           if 'somefield' in myform:
               print("my form has some field")
        """
        return bool(self.fields and name in self.fields)

    def __iter__(self):
        """
        Custom logic to allow iterating over form field names::

           myform = Form(fields=['foo', 'bar'])
           for fieldname in myform:
               print(fieldname)
        """
        return iter(self.fields)

    @property
    def vue_component(self):
        """
        String name for the Vue component, e.g. ``'WuttaForm'``.

        This is a generated value based on :attr:`vue_tagname`.
        """
        words = self.vue_tagname.split('-')
        return ''.join([word.capitalize() for word in words])

    def get_cancel_url(self):
        """
        Returns the URL for the Cancel button.

        If :attr:`cancel_url` is set, its value is returned.

        Or, if the referrer can be deduced from the request, that is
        returned.

        Or, if :attr:`cancel_url_fallback` is set, that value is
        returned.

        As a last resort the "default" URL from
        :func:`~wuttaweb.subscribers.request.get_referrer()` is
        returned.
        """
        # use "permanent" URL if set
        if self.cancel_url:
            return self.cancel_url

        # nb. use fake default to avoid normal default logic;
        # that way if we get something it's a real referrer
        url = self.request.get_referrer(default='NOPE')
        if url and url != 'NOPE':
            return url

        # use fallback URL if set
        if self.cancel_url_fallback:
            return self.cancel_url_fallback

        # okay, home page then (or whatever is the default URL)
        return self.request.get_referrer()

    def set_fields(self, fields):
        """
        Explicitly set the list of form fields.

        This will overwrite :attr:`fields` with a new
        :class:`~wuttaweb.util.FieldList` instance.

        :param fields: List of string field names.
        """
        self.fields = FieldList(fields)

    def append(self, *keys):
        """
        Add some fields(s) to the form.

        This is a convenience to allow adding multiple fields at
        once::

           form.append('first_field',
                       'second_field',
                       'third_field')

        It will add each field to :attr:`fields`.
        """
        for key in keys:
            if key not in self.fields:
                self.fields.append(key)

    def remove(self, *keys):
        """
        Remove some fields(s) from the form.

        This is a convenience to allow removal of multiple fields at
        once::

           form.remove('first_field',
                       'second_field',
                       'third_field')

        It will remove each field from :attr:`fields`.
        """
        for key in keys:
            if key in self.fields:
                self.fields.remove(key)

    def set_node(self, key, nodeinfo, **kwargs):
        """
        Set/override the node for a field.

        :param key: Name of field.

        :param nodeinfo: Should be either a
           :class:`colander:colander.SchemaNode` instance, or else a
           :class:`colander:colander.SchemaType` instance.

        If ``nodeinfo`` is a proper node instance, it will be used
        as-is.  Otherwise an
        :class:`~wuttaweb.forms.schema.ObjectNode` instance will be
        constructed using ``nodeinfo`` as the type (``typ``).

        Node overrides are tracked via :attr:`nodes`.
        """
        from wuttaweb.forms.schema import ObjectNode

        if isinstance(nodeinfo, colander.SchemaNode):
            # assume nodeinfo is a complete node
            node = nodeinfo

        else: # assume nodeinfo is a schema type
            kwargs.setdefault('name', key)
            node = ObjectNode(nodeinfo, **kwargs)

        self.nodes[key] = node

        # must explicitly replace node, if we already have a schema
        if self.schema:
            self.schema[key] = node

    def set_widget(self, key, widget, **kwargs):
        """
        Set/override the widget for a field.

        You can specify a widget instance or else a named "type" of
        widget, in which case that is passed along to
        :meth:`make_widget()`.

        :param key: Name of field.

        :param widget: Either a :class:`deform:deform.widget.Widget`
           instance, or else a widget "type" name.

        :param \**kwargs: Any remaining kwargs are passed along to
           :meth:`make_widget()` - if applicable.

        Widget overrides are tracked via :attr:`widgets`.
        """
        if not isinstance(widget, deform.widget.Widget):
            widget_obj = self.make_widget(widget, **kwargs)
            if not widget_obj:
                raise ValueError(f"widget type not supported: {widget}")
            widget = widget_obj

        self.widgets[key] = widget

        # update schema if necessary
        if self.schema and key in self.schema:
            self.schema[key].widget = widget

    def make_widget(self, widget_type, **kwargs):
        """
        Make and return a new field widget of the given type.

        This has built-in support for the following types (although
        subclass can override as needed):

        * ``'notes'`` => :class:`~wuttaweb.forms.widgets.NotesWidget`

        See also :meth:`set_widget()` which may call this method
        automatically.

        :param widget_type: Which of the above (or custom) widget
           type to create.

        :param \**kwargs: Remaining kwargs are passed as-is to the
           widget factory.

        :returns: New widget instance, or ``None`` if e.g. it could
           not determine how to create the widget.
        """
        from wuttaweb.forms import widgets

        if widget_type == 'notes':
            return widgets.NotesWidget(**kwargs)

    def set_default_widgets(self):
        """
        Set default field widgets, where applicable.

        This will add new entries to :attr:`widgets` for columns
        whose data type implies a default widget should be used.
        This is generally only possible if :attr:`model_class` is set
        to a valid SQLAlchemy mapped class.

        This only checks for a couple of data types, with mapping as
        follows:

        * :class:`sqlalchemy:sqlalchemy.types.Date` ->
          :class:`~wuttaweb.forms.widgets.WuttaDateWidget`
        * :class:`sqlalchemy:sqlalchemy.types.DateTime` ->
          :class:`~wuttaweb.forms.widgets.WuttaDateTimeWidget`
        """
        from wuttaweb.forms import widgets

        if not self.model_class:
            return

        for key in self.fields:
            if key in self.widgets:
                continue

            attr = getattr(self.model_class, key, None)
            if attr:
                prop = getattr(attr, 'prop', None)
                if prop and isinstance(prop, orm.ColumnProperty):
                    column = prop.columns[0]
                    if isinstance(column.type, sa.Date):
                        self.set_widget(key, widgets.WuttaDateWidget(self.request))
                    elif isinstance(column.type, sa.DateTime):
                        self.set_widget(key, widgets.WuttaDateTimeWidget(self.request))

    def set_grid(self, key, grid):
        """
        Establish a :term:`grid` to be displayed for a field.  This
        uses a :class:`~wuttaweb.forms.widgets.GridWidget` to wrap the
        rendered grid.

        :param key: Name of field.

        :param widget: :class:`~wuttaweb.grids.base.Grid` instance,
           pre-configured and (usually) with data.
        """
        from wuttaweb.forms.widgets import GridWidget

        widget = GridWidget(self.request, grid)
        self.set_widget(key, widget)
        self.add_grid_vue_context(grid)

    def add_grid_vue_context(self, grid):
        """ """
        if not grid.key:
            raise ValueError("grid must have a key!")

        if grid.key in self.grid_vue_context:
            log.warning("grid data with key '%s' already registered, "
                        "but will be replaced", grid.key)

        self.grid_vue_context[grid.key] = grid.get_vue_context()

    def set_validator(self, key, validator):
        """
        Set/override the validator for a field, or the form.

        :param key: Name of field.  This may also be ``None`` in which
           case the validator will apply to the whole form instead of
           a field.

        :param validator: Callable which accepts ``(node, value)``
           args.  For instance::

              def validate_foo(node, value):
                  if value == 42:
                      node.raise_invalid("42 is not allowed!")

              form = Form(fields=['foo', 'bar'])

              form.set_validator('foo', validate_foo)

        Validator overrides are tracked via :attr:`validators`.
        """
        self.validators[key] = validator

        # nb. must apply to existing schema if present
        if self.schema and key in self.schema:
            self.schema[key].validator = validator

    def set_default(self, key, value):
        """
        Set/override the default value for a field.

        :param key: Name of field.

        :param validator: Default value for the field.

        Default value overrides are tracked via :attr:`defaults`.
        """
        self.defaults[key] = value

    def set_readonly(self, key, readonly=True):
        """
        Enable or disable the "readonly" flag for a given field.

        When a field is marked readonly, it will be shown in the form
        but there will be no editable widget.  The field is skipped
        over (not saved) when form is submitted.

        See also :meth:`is_readonly()`; this is tracked via
        :attr:`readonly_fields`.

        :param key: String key (fieldname) for the field.

        :param readonly: New readonly flag for the field.
        """
        if readonly:
            self.readonly_fields.add(key)
        else:
            if key in self.readonly_fields:
                self.readonly_fields.remove(key)

    def is_readonly(self, key):
        """
        Returns boolean indicating if the given field is marked as
        readonly.

        See also :meth:`set_readonly()`.

        :param key: Field key/name as string.
        """
        if self.readonly_fields:
            if key in self.readonly_fields:
                return True
        return False

    def set_required(self, key, required=True):
        """
        Enable or disable the "required" flag for a given field.

        When a field is marked required, a value must be provided
        or else it fails validation.

        In practice if a field is "not required" then a default
        "empty" value is assumed, should the user not provide one.

        See also :meth:`is_required()`; this is tracked via
        :attr:`required_fields`.

        :param key: String key (fieldname) for the field.

        :param required: New required flag for the field.  Usually a
           boolean, but may also be ``None`` to remove any flag and
           revert to default behavior for the field.
        """
        self.required_fields[key] = required

    def is_required(self, key):
        """
        Returns boolean indicating if the given field is marked as
        required.

        See also :meth:`set_required()`.

        :param key: Field key/name as string.

        :returns: Value for the flag from :attr:`required_fields` if
           present; otherwise ``None``.
        """
        return self.required_fields.get(key, None)

    def set_label(self, key, label):
        """
        Set the label for given field name.

        See also :meth:`get_label()`.
        """
        self.labels[key] = label

        # update schema if necessary
        if self.schema and key in self.schema:
            self.schema[key].title = label

    def get_label(self, key):
        """
        Get the label for given field name.

        Note that this will always return a string, auto-generating
        the label if needed.

        See also :meth:`set_label()`.
        """
        return self.labels.get(key, self.app.make_title(key))

    def get_fields(self):
        """
        Returns the official list of field names for the form, or
        ``None``.

        If :attr:`fields` is set and non-empty, it is returned.

        Or, if :attr:`schema` is set, the field list is derived
        from that.

        Or, if :attr:`model_class` is set, the field list is derived
        from that, via :meth:`get_model_fields()`.

        Otherwise ``None`` is returned.
        """
        if hasattr(self, 'fields') and self.fields:
            return self.fields

        if self.schema:
            return [field.name for field in self.schema]

        fields = self.get_model_fields()
        if fields:
            return fields

        return []

    def get_model_fields(self, model_class=None):
        """
        This method is a shortcut which calls
        :func:`~wuttaweb.util.get_model_fields()`.

        :param model_class: Optional model class for which to return
           fields.  If not set, the form's :attr:`model_class` is
           assumed.
        """
        return get_model_fields(self.config,
                                model_class=model_class or self.model_class)

    def get_schema(self):
        """
        Return the :class:`colander:colander.Schema` object for the
        form, generating it automatically if necessary.

        Note that if :attr:`schema` is already set, that will be
        returned as-is.
        """
        if not self.schema:

            ##############################
            # create schema
            ##############################

            # get fields
            fields = self.get_fields()
            if not fields:
                raise NotImplementedError

            if self.model_class:

                # collect list of field names and/or nodes
                includes = []
                for key in fields:
                    if key in self.nodes:
                        includes.append(self.nodes[key])
                    else:
                        includes.append(key)

                # make initial schema with ColanderAlchemy magic
                schema = SQLAlchemySchemaNode(self.model_class,
                                              includes=includes)

                # fill in the blanks if anything got missed
                for key in fields:
                    if key not in schema:
                        node = colander.SchemaNode(colander.String(), name=key)
                        schema.add(node)

            else:

                # make basic schema
                schema = colander.Schema()
                for key in fields:
                    node = None

                    # use node override if present
                    if key in self.nodes:
                        node = self.nodes[key]
                    if not node:

                        # otherwise make simple string node
                        node = colander.SchemaNode(
                            colander.String(),
                            name=key)

                    schema.add(node)

            ##############################
            # customize schema
            ##############################

            # apply widget overrides
            for key, widget in self.widgets.items():
                if key in schema:
                    schema[key].widget = widget

            # apply validator overrides
            for key, validator in self.validators.items():
                if key is None:
                    # nb. this one is form-wide
                    schema.validator = validator
                elif key in schema: # field-level
                    schema[key].validator = validator

            # apply default value overrides
            for key, value in self.defaults.items():
                if key in schema:
                    schema[key].default = value

            # apply required flags
            for key, required in self.required_fields.items():
                if key in schema:
                    if required is False:
                        schema[key].missing = colander.null

            self.schema = schema

        return self.schema

    def get_deform(self):
        """
        Return the :class:`deform:deform.Form` instance for the form,
        generating it automatically if necessary.
        """
        if not hasattr(self, 'deform_form'):
            model = self.app.model
            schema = self.get_schema()
            kwargs = {}

            if self.model_instance:

                # TODO: i keep finding problems with this, not sure
                # what needs to happen.  some forms will have a simple
                # dict for model_instance, others will have a proper
                # SQLAlchemy object.  and in the latter case, it may
                # not be "wutta-native" but from another DB.

                # so the problem is, how to detect whether we should
                # use the model_instance as-is or if we should convert
                # to a dict.  some options include:

                # - check if instance has dictify() method
                # i *think* this was tried and didn't work? but do not recall

                # - check if is instance of model.Base
                # this is unreliable since model.Base is wutta-native

                # - check if form has a model_class
                # has not been tried yet

                # - check if schema is from colanderalchemy
                # this is what we are trying currently...

                if isinstance(schema, SQLAlchemySchemaNode):
                    kwargs['appstruct'] = schema.dictify(self.model_instance)
                else:
                    kwargs['appstruct'] = self.model_instance

            # create the Deform instance
            # nb. must give a reference back to wutta form; this is
            # for sake of field schema nodes and widgets, e.g. to
            # access the main model instance
            form = deform.Form(schema, **kwargs)
            form.wutta_form = self
            self.deform_form = form

        return self.deform_form

    def render_vue_tag(self, **kwargs):
        """
        Render the Vue component tag for the form.

        By default this simply returns:

        .. code-block:: html

           <wutta-form></wutta-form>

        The actual output will depend on various form attributes, in
        particular :attr:`vue_tagname`.
        """
        return HTML.tag(self.vue_tagname, **kwargs)

    def render_vue_template(
            self,
            template='/forms/vue_template.mako',
            **context):
        """
        Render the Vue template block for the form.

        This returns something like:

        .. code-block:: none

           <script type="text/x-template" id="wutta-form-template">
             <form>
               <!-- fields etc. -->
             </form>
           </script>

           <script>
               WuttaFormData = {}
               WuttaForm = {
                   template: 'wutta-form-template',
               }
           </script>

        .. todo::

           Why can't Sphinx render the above code block as 'html' ?

           It acts like it can't handle a ``<script>`` tag at all?

        Actual output will of course depend on form attributes, i.e.
        :attr:`vue_tagname` and :attr:`fields` list etc.

        :param template: Path to Mako template which is used to render
           the output.
        """
        context['form'] = self
        context['dform'] = self.get_deform()
        context.setdefault('request', self.request)
        context['model_data'] = self.get_vue_model_data()

        # set form method, enctype
        context.setdefault('form_attrs', {})
        context['form_attrs'].setdefault('method', self.action_method)
        if self.action_method == 'post':
            context['form_attrs'].setdefault('enctype', 'multipart/form-data')

        # auto disable button on submit
        if self.auto_disable_submit:
            context['form_attrs']['@submit'] = 'formSubmitting = true'

        output = render(template, context)
        return HTML.literal(output)

    def render_vue_field(
            self,
            fieldname,
            readonly=None,
            **kwargs,
    ):
        """
        Render the given field completely, i.e. ``<b-field>`` wrapper
        with label and containing a widget.

        Actual output will depend on the field attributes etc.
        Typical output might look like:

        .. code-block:: html

           <b-field label="Foo"
                    horizontal
                    type="is-danger"
                    message="something went wrong!">
              <!-- widget element(s) -->
           </b-field>

        .. warning::

           Any ``**kwargs`` received from caller are ignored by this
           method.  For now they are allowed, for sake of backwawrd
           compatibility.  This may change in the future.
        """
        # readonly comes from: caller, field flag, or form flag
        if readonly is None:
            readonly = self.is_readonly(fieldname)
            if not readonly:
                readonly = self.readonly

        # but also, fields not in deform/schema must be readonly
        dform = self.get_deform()
        if not readonly and fieldname not in dform:
            readonly = True

        # render the field widget or whatever
        if fieldname in dform:

            # render proper widget if field is in deform/schema
            field = dform[fieldname]
            kw = {}
            if readonly:
                kw['readonly'] = True
            html = field.serialize(**kw)

        else:
            # render static text if field not in deform/schema
            # TODO: need to abstract this somehow
            if self.model_instance:
                html = str(self.model_instance[fieldname])
            else:
                html = ''

        # mark all that as safe
        html = HTML.literal(html)

        # render field label
        label = self.get_label(fieldname)

        # b-field attrs
        attrs = {
            ':horizontal': 'true',
            'label': label,
        }

        # next we will build array of messages to display..some
        # fields always show a "helptext" msg, and some may have
        # validation errors..
        field_type = None
        messages = []

        # show errors if present
        errors = self.get_field_errors(fieldname)
        if errors:
            field_type = 'is-danger'
            messages.extend(errors)

        # ..okay now we can declare the field messages and type
        if field_type:
            attrs['type'] = field_type
        if messages:
            cls = 'is-size-7'
            if field_type == 'is-danger':
                cls += ' has-text-danger'
            messages = [HTML.tag('p', c=[msg], class_=cls)
                        for msg in messages]
            slot = HTML.tag('slot', name='messages', c=messages)
            html = HTML.tag('div', c=[html, slot])

        return HTML.tag('b-field', c=[html], **attrs)

    def render_vue_finalize(self):
        """
        Render the Vue "finalize" script for the form.

        By default this simply returns:

        .. code-block:: html

           <script>
             WuttaForm.data = function() { return WuttaFormData }
             Vue.component('wutta-form', WuttaForm)
           </script>

        The actual output may depend on various form attributes, in
        particular :attr:`vue_tagname`.
        """
        set_data = f"{self.vue_component}.data = function() {{ return {self.vue_component}Data }}"
        make_component = f"Vue.component('{self.vue_tagname}', {self.vue_component})"
        return HTML.tag('script', c=['\n',
                                     HTML.literal(set_data),
                                     '\n',
                                     HTML.literal(make_component),
                                     '\n'])

    def get_vue_model_data(self):
        """
        Returns a dict with form model data.  Values may be nested
        depending on the types of fields contained in the form.

        This collects the ``cstruct`` values for all fields which are
        present both in :attr:`fields` as well as the Deform schema.

        It also converts each as needed, to ensure it is
        JSON-serializable.

        :returns: Dict of field/value items.
        """
        dform = self.get_deform()
        model_data = {}

        def assign(field):
            value = field.cstruct

            # TODO: we need a proper true/false on the Vue side,
            # but deform/colander want 'true' and 'false' ..so
            # for now we explicitly translate here, ugh.  also
            # note this does not yet allow for null values.. :(
            if isinstance(field.typ, colander.Boolean):
                value = True if value == field.typ.true_val else False

            model_data[field.oid] = make_json_safe(value)

        for key in self.fields:

            # TODO: i thought commented code was useful, but no longer sure?

            # TODO: need to describe the scenario when this is true
            if key not in dform:
                # log.warning("field '%s' is missing from deform", key)
                continue

            field = dform[key]

            # if hasattr(field, 'children'):
            #     for subfield in field.children:
            #         assign(subfield)

            assign(field)

        return model_data

    # TODO: for tailbone compat, should document?
    # (ideally should remove this and find a better way)
    def get_vue_field_value(self, key):
        """ """
        if key not in self.fields:
            return

        dform = self.get_deform()
        if key not in dform:
            return

        field = dform[key]
        return make_json_safe(field.cstruct)

    def validate(self):
        """
        Try to validate the form, using data from the :attr:`request`.

        Uses :func:`~wuttaweb.util.get_form_data()` to retrieve the
        form data from POST or JSON body.

        If the form data is valid, the data dict is returned.  This
        data dict is also made available on the form object via the
        :attr:`validated` attribute.

        However if the data is not valid, ``False`` is returned, and
        there will be no :attr:`validated` attribute.  In that case
        you should inspect the form errors to learn/display what went
        wrong for the user's sake.  See also
        :meth:`get_field_errors()`.

        This uses :meth:`deform:deform.Field.validate()` under the
        hood.

        .. warning::

           Calling ``validate()`` on some forms will cause the
           underlying Deform and Colander structures to mutate.  In
           particular, all :attr:`readonly_fields` will be *removed*
           from the :attr:`schema` to ensure they are not involved in
           the validation.

        :returns: Data dict, or ``False``.
        """
        if hasattr(self, 'validated'):
            del self.validated

        if self.request.method != 'POST':
            return False

        # remove all readonly fields from deform / schema
        dform = self.get_deform()
        if self.readonly_fields:
            schema = self.get_schema()
            for field in self.readonly_fields:
                if field in schema:
                    del schema[field]
                    dform.children.remove(dform[field])

        # let deform do real validation
        controls = get_form_data(self.request).items()
        try:
            self.validated = dform.validate(controls)
        except deform.ValidationFailure:
            log.debug("form not valid: %s", dform.error)
            return False

        return self.validated

    def has_global_errors(self):
        """
        Convenience function to check if the form has any "global"
        (not field-level) errors.

        See also :meth:`get_global_errors()`.

        :returns: ``True`` if global errors present, else ``False``.
        """
        dform = self.get_deform()
        return bool(dform.error)

    def get_global_errors(self):
        """
        Returns a list of "global" (not field-level) error messages
        for the form.

        See also :meth:`has_global_errors()`.

        :returns: List of error messages (possibly empty).
        """
        dform = self.get_deform()
        if dform.error is None:
            return []
        return dform.error.messages()

    def get_field_errors(self, field):
        """
        Return a list of error messages for the given field.

        Not useful unless a call to :meth:`validate()` failed.
        """
        dform = self.get_deform()
        if field in dform:
            field = dform[field]
            if field.error:
                return field.error.messages()
        return []



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
SQLAlchemy-Continuum Plugin
"""

from pyramid.threadlocal import get_current_request

try:
    from wutta_continuum.conf import WuttaContinuumPlugin
except ImportError: # pragma: no cover
    pass
else:

    class WuttaWebContinuumPlugin(WuttaContinuumPlugin):
        """
        SQLAlchemy-Continuum manager plugin for WuttaWeb.

        This tries to use the current request to obtain user and IP
        address for the transaction.
        """

        # TODO: should find a better way, threadlocals are bad?
        # https://docs.pylonsproject.org/projects/pyramid/en/latest/api/threadlocal.html#pyramid.threadlocal.get_current_request

        def get_remote_addr(self, uow, session):
            """ """
            request = get_current_request()
            if request:
                return request.client_addr

        def get_user_id(self, uow, session):
            """ """
            request = get_current_request()
            if request and request.user:
                return request.user.uuid



# -*- coding: utf-8; -*-

import datetime
import decimal
from unittest import TestCase
from unittest.mock import patch, MagicMock

import sqlalchemy as sa
from sqlalchemy import orm
from paginate import Page
from paginate_sqlalchemy import SqlalchemyOrmPage
from pyramid import testing

from wuttjamaican.conf import WuttaConfig
from wuttaweb.grids import base as mod
from wuttaweb.grids.filters import GridFilter, StringAlchemyFilter, default_sqlalchemy_filters
from wuttaweb.util import FieldList
from wuttaweb.forms import Form
from wuttaweb.testing import WebTestCase


class TestGrid(WebTestCase):

    def make_grid(self, request=None, **kwargs):
        return mod.Grid(request or self.request, **kwargs)

    def test_constructor(self):

        # empty
        grid = self.make_grid()
        self.assertIsNone(grid.key)
        self.assertEqual(grid.columns, [])
        self.assertIsNone(grid.data)

        # now with columns
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertIsInstance(grid.columns, FieldList)
        self.assertEqual(grid.columns, ['foo', 'bar'])

    def test_constructor_sorting(self):
        model = self.app.model

        # defaults, not sortable
        grid = self.make_grid()
        self.assertFalse(grid.sortable)
        self.assertTrue(grid.sort_on_backend)
        self.assertEqual(grid.sorters, {})
        self.assertEqual(grid.sort_defaults, [])

        # defaults, sortable
        grid = self.make_grid(sortable=True)
        self.assertTrue(grid.sortable)
        self.assertTrue(grid.sort_on_backend)
        self.assertEqual(grid.sorters, {})
        self.assertEqual(grid.sort_defaults, [])

        # sorters may be pre-populated
        grid = self.make_grid(model_class=model.Setting, sortable=True)
        self.assertEqual(len(grid.sorters), 2)
        self.assertIn('name', grid.sorters)
        self.assertIn('value', grid.sorters)
        self.assertEqual(grid.sort_defaults, [])

        # sort defaults as str
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_defaults='name')
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'asc')])

        # sort defaults as tuple
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_defaults=('name', 'desc'))
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'desc')])

        # sort defaults as list w/ single tuple
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_defaults=[('name', 'desc')])
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'desc')])

        # multi-column defaults
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_multiple=True,
                              sort_defaults=[('name', 'desc'), ('value', 'asc')])
        self.assertTrue(grid.sort_multiple)
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'desc'),
                                              mod.SortInfo('value', 'asc')])

        # multi-column sort disabled for oruga
        self.request.use_oruga = True
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_multiple=True)
        self.assertFalse(grid.sort_multiple)

    def test_constructor_filtering(self):
        model = self.app.model

        # defaults, not filterable
        grid = self.make_grid()
        self.assertFalse(grid.filterable)
        self.assertEqual(grid.filters, {})

        # defaults, filterable
        grid = self.make_grid(filterable=True)
        self.assertTrue(grid.filterable)
        self.assertEqual(grid.filters, {})

        # filters may be pre-populated
        with patch.object(mod.Grid, 'make_filter', return_value=42):
            grid = self.make_grid(model_class=model.Setting, filterable=True)
            self.assertEqual(len(grid.filters), 2)
            self.assertIn('name', grid.filters)
            self.assertIn('value', grid.filters)

        # can specify filters
        grid = self.make_grid(model_class=model.Setting, filterable=True,
                              filters={'name': 42})
        self.assertTrue(grid.filterable)
        self.assertEqual(grid.filters, {'name': 42})

    def test_vue_tagname(self):
        grid = self.make_grid()
        self.assertEqual(grid.vue_tagname, 'wutta-grid')

    def test_vue_component(self):
        grid = self.make_grid()
        self.assertEqual(grid.vue_component, 'WuttaGrid')

    def test_get_columns(self):
        model = self.app.model

        # empty
        grid = self.make_grid()
        self.assertEqual(grid.columns, [])
        self.assertEqual(grid.get_columns(), [])

        # explicit
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.columns, ['foo', 'bar'])
        self.assertEqual(grid.get_columns(), ['foo', 'bar'])

        # derived from model
        grid = self.make_grid(model_class=model.Setting)
        self.assertEqual(grid.columns, ['name', 'value'])
        self.assertEqual(grid.get_columns(), ['name', 'value'])

    def test_append(self):
        grid = self.make_grid(columns=['one', 'two'])
        self.assertEqual(grid.columns, ['one', 'two'])
        grid.append('one', 'two', 'three')
        self.assertEqual(grid.columns, ['one', 'two', 'three'])

    def test_remove(self):
        grid = self.make_grid(columns=['one', 'two', 'three', 'four'])
        self.assertEqual(grid.columns, ['one', 'two', 'three', 'four'])
        grid.remove('two', 'three')
        self.assertEqual(grid.columns, ['one', 'four'])

    def test_set_label(self):
        model = self.app.model
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.Setting, filterable=True)
        self.assertEqual(grid.labels, {})

        # basic
        grid.set_label('name', "NAME COL")
        self.assertEqual(grid.labels['name'], "NAME COL")

        # can replace label
        grid.set_label('name', "Different")
        self.assertEqual(grid.labels['name'], "Different")
        self.assertEqual(grid.get_label('name'), "Different")

        # can update only column, not filter
        self.assertEqual(grid.labels, {'name': "Different"})
        self.assertIn('name', grid.filters)
        self.assertEqual(grid.filters['name'].label, "Different")
        grid.set_label('name', "COLUMN ONLY", column_only=True)
        self.assertEqual(grid.get_label('name'), "COLUMN ONLY")
        self.assertEqual(grid.filters['name'].label, "Different")

    def test_get_label(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.labels, {})

        # default derived from key
        self.assertEqual(grid.get_label('foo'), "Foo")

        # can override
        grid.set_label('foo', "Different")
        self.assertEqual(grid.get_label('foo'), "Different")

    def test_set_renderer(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.renderers, {})

        def render1(record, key, value):
            pass

        # basic
        grid.set_renderer('foo', render1)
        self.assertIs(grid.renderers['foo'], render1)

        def render2(record, key, value, extra=None):
            return extra

        # can pass kwargs to get a partial
        grid.set_renderer('foo', render2, extra=42)
        self.assertIsNot(grid.renderers['foo'], render2)
        self.assertEqual(grid.renderers['foo'](None, None, None), 42)

        # can use built-in string shortcut
        grid.set_renderer('foo', 'quantity')
        obj = MagicMock(foo=42.00)
        self.assertEqual(grid.renderers['foo'](obj, 'foo', 42.00), '42')

    def test_set_default_renderers(self):
        model = self.app.model

        # no defaults for "plain" schema
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.renderers, {})

        # no defaults for "plain" mapped class
        grid = self.make_grid(model_class=model.Setting)
        self.assertEqual(grid.renderers, {})

        def myrender(obj, key, value):
            return value

        # renderer set for datetime mapped field
        grid = self.make_grid(model_class=model.Upgrade)
        self.assertIn('created', grid.renderers)
        self.assertIsNot(grid.renderers['created'], myrender)

        # renderer *not* set for datetime, if override present
        grid = self.make_grid(model_class=model.Upgrade,
                              renderers={'created': myrender})
        self.assertIn('created', grid.renderers)
        self.assertIs(grid.renderers['created'], myrender)

        # renderer set for boolean mapped field
        grid = self.make_grid(model_class=model.Upgrade)
        self.assertIn('executing', grid.renderers)
        self.assertIsNot(grid.renderers['executing'], myrender)

        # renderer *not* set for boolean, if override present
        grid = self.make_grid(model_class=model.Upgrade,
                              renderers={'executing': myrender})
        self.assertIn('executing', grid.renderers)
        self.assertIs(grid.renderers['executing'], myrender)

        # nb. as of writing we have no Date columns in default schema,
        # so must invent one to test that type
        class SomeFoolery(model.Base):
            __tablename__ = 'somefoolery'
            id = sa.Column(sa.Integer(), primary_key=True)
            created = sa.Column(sa.Date())

        # renderer set for date mapped field
        grid = self.make_grid(model_class=SomeFoolery)
        self.assertIn('created', grid.renderers)
        self.assertIsNot(grid.renderers['created'], myrender)

    def test_linked_columns(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.linked_columns, [])
        self.assertFalse(grid.is_linked('foo'))

        grid.set_link('foo')
        self.assertEqual(grid.linked_columns, ['foo'])
        self.assertTrue(grid.is_linked('foo'))
        self.assertFalse(grid.is_linked('bar'))

        grid.set_link('bar')
        self.assertEqual(grid.linked_columns, ['foo', 'bar'])
        self.assertTrue(grid.is_linked('foo'))
        self.assertTrue(grid.is_linked('bar'))

        grid.set_link('foo', False)
        self.assertEqual(grid.linked_columns, ['bar'])
        self.assertFalse(grid.is_linked('foo'))
        self.assertTrue(grid.is_linked('bar'))

    def test_searchable_columns(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        self.assertEqual(grid.searchable_columns, set())
        self.assertFalse(grid.is_searchable('foo'))

        grid.set_searchable('foo')
        self.assertEqual(grid.searchable_columns, {'foo'})
        self.assertTrue(grid.is_searchable('foo'))
        self.assertFalse(grid.is_searchable('bar'))

        grid.set_searchable('bar')
        self.assertEqual(grid.searchable_columns, {'foo', 'bar'})
        self.assertTrue(grid.is_searchable('foo'))
        self.assertTrue(grid.is_searchable('bar'))

        grid.set_searchable('foo', False)
        self.assertEqual(grid.searchable_columns, {'bar'})
        self.assertFalse(grid.is_searchable('foo'))
        self.assertTrue(grid.is_searchable('bar'))

    def test_add_action(self):
        grid = self.make_grid()
        self.assertEqual(len(grid.actions), 0)

        grid.add_action('view')
        self.assertEqual(len(grid.actions), 1)
        self.assertIsInstance(grid.actions[0], mod.GridAction)

    def test_set_tools(self):
        grid = self.make_grid()
        self.assertEqual(grid.tools, {})

        # null
        grid.set_tools(None)
        self.assertEqual(grid.tools, {})

        # empty
        grid.set_tools({})
        self.assertEqual(grid.tools, {})

        # full dict is replaced
        grid.tools = {'foo': 'bar'}
        self.assertEqual(grid.tools, {'foo': 'bar'})
        grid.set_tools({'bar': 'baz'})
        self.assertEqual(grid.tools, {'bar': 'baz'})

        # can specify as list of html elements
        grid.set_tools(['foo', 'bar'])
        self.assertEqual(len(grid.tools), 2)
        self.assertEqual(list(grid.tools.values()), ['foo', 'bar'])

    def test_add_tool(self):
        grid = self.make_grid()
        self.assertEqual(grid.tools, {})

        # with key
        grid.add_tool('foo', key='foo')
        self.assertEqual(grid.tools, {'foo': 'foo'})

        # without key
        grid.add_tool('bar')
        self.assertEqual(len(grid.tools), 2)
        self.assertEqual(list(grid.tools.values()), ['foo', 'bar'])

    def test_get_pagesize_options(self):
        grid = self.make_grid()

        # default
        options = grid.get_pagesize_options()
        self.assertEqual(options, [5, 10, 20, 50, 100, 200])

        # override default
        options = grid.get_pagesize_options(default=[42])
        self.assertEqual(options, [42])

        # from config
        self.config.setdefault('wuttaweb.grids.default_pagesize_options', '1 2 3')
        options = grid.get_pagesize_options()
        self.assertEqual(options, [1, 2, 3])

    def test_get_pagesize(self):
        grid = self.make_grid()

        # default
        size = grid.get_pagesize()
        self.assertEqual(size, 20)

        # override default
        size = grid.get_pagesize(default=42)
        self.assertEqual(size, 42)

        # override default options
        self.config.setdefault('wuttaweb.grids.default_pagesize_options', '10 15 30')
        grid = self.make_grid()
        size = grid.get_pagesize()
        self.assertEqual(size, 10)

        # from config
        self.config.setdefault('wuttaweb.grids.default_pagesize', '15')
        size = grid.get_pagesize()
        self.assertEqual(size, 15)

    ##############################
    # configuration methods
    ##############################

    def test_load_settings(self):
        model = self.app.model

        # nb. first use a paging grid
        grid = self.make_grid(key='foo', paginated=True, paginate_on_backend=True,
                              pagesize=20, page=1)

        # settings are loaded, applied, saved
        self.assertEqual(grid.page, 1)
        self.assertNotIn('grid.foo.page', self.request.session)
        self.request.GET = {'pagesize': '10', 'page': '2'}
        grid.load_settings()
        self.assertEqual(grid.page, 2)
        self.assertEqual(self.request.session['grid.foo.page'], 2)

        # can skip the saving step
        self.request.GET = {'pagesize': '10', 'page': '3'}
        grid.load_settings(persist=False)
        self.assertEqual(grid.page, 3)
        self.assertEqual(self.request.session['grid.foo.page'], 2)

        # no error for non-paginated grid
        grid = self.make_grid(key='foo', paginated=False)
        grid.load_settings()
        self.assertFalse(grid.paginated)

        # nb. next use a sorting grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True)

        # settings are loaded, applied, saved
        self.assertEqual(grid.sort_defaults, [])
        self.assertFalse(hasattr(grid, 'active_sorters'))
        self.request.GET = {'sort1key': 'name', 'sort1dir': 'desc'}
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'desc'}])
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 1)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'desc')

        # can skip the saving step
        self.request.GET = {'sort1key': 'name', 'sort1dir': 'asc'}
        grid.load_settings(persist=False)
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'asc'}])
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 1)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'desc')

        # no error for non-sortable grid
        grid = self.make_grid(key='foo', sortable=False)
        grid.load_settings()
        self.assertFalse(grid.sortable)

        # with sort defaults
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_on_backend=True, sort_defaults='name')
        self.assertFalse(hasattr(grid, 'active_sorters'))
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'asc'}])

        # with multi-column sort defaults
        grid = self.make_grid(model_class=model.Setting, sortable=True,
                              sort_on_backend=True)
        grid.sort_defaults = [
            mod.SortInfo('name', 'asc'),
            mod.SortInfo('value', 'desc'),
        ]
        self.assertFalse(hasattr(grid, 'active_sorters'))
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'asc'}])

        # load settings from session when nothing is in request
        self.request.GET = {}
        self.request.session.invalidate()
        self.assertNotIn('grid.settings.sorters.length', self.request.session)
        self.request.session['grid.settings.sorters.length'] = 1
        self.request.session['grid.settings.sorters.1.key'] = 'name'
        self.request.session['grid.settings.sorters.1.dir'] = 'desc'
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True,
                              paginated=True, paginate_on_backend=True)
        self.assertFalse(hasattr(grid, 'active_sorters'))
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'desc'}])

        # filter settings are loaded, applied, saved
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              filterable=True)
        self.assertEqual(len(grid.filters), 2)
        self.assertEqual(len(grid.active_filters), 0)
        self.assertNotIn('grid.settings.filter.name.active', self.request.session)
        self.assertNotIn('grid.settings.filter.value.active', self.request.session)
        self.request.GET = {'name': 'john', 'name.verb': 'contains'}
        grid.load_settings()
        self.assertTrue(grid.filters['name'].active)
        self.assertEqual(grid.filters['name'].verb, 'contains')
        self.assertEqual(grid.filters['name'].value, 'john')
        self.assertTrue(self.request.session['grid.settings.filter.name.active'])
        self.assertEqual(self.request.session['grid.settings.filter.name.verb'], 'contains')
        self.assertEqual(self.request.session['grid.settings.filter.name.value'], 'john')

        # filter + sort settings are loaded, applied, saved
        self.request.session.invalidate()
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, filterable=True)
        self.assertEqual(len(grid.filters), 2)
        self.assertEqual(len(grid.active_filters), 0)
        self.assertNotIn('grid.settings.filter.name.active', self.request.session)
        self.assertNotIn('grid.settings.filter.value.active', self.request.session)
        self.assertNotIn('grid.settings.sorters.length', self.request.session)
        self.request.GET = {'name': 'john', 'name.verb': 'contains',
                            'sort1key': 'name', 'sort1dir': 'asc'}
        grid.load_settings()
        self.assertTrue(grid.filters['name'].active)
        self.assertEqual(grid.filters['name'].verb, 'contains')
        self.assertEqual(grid.filters['name'].value, 'john')
        self.assertTrue(self.request.session['grid.settings.filter.name.active'])
        self.assertEqual(self.request.session['grid.settings.filter.name.verb'], 'contains')
        self.assertEqual(self.request.session['grid.settings.filter.name.value'], 'john')
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 1)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'asc')

        # can reset view to defaults
        self.request.GET = {'reset-view': 'true'}
        grid.load_settings()
        self.assertEqual(grid.active_filters, [])
        self.assertIsNone(grid.filters['name'].value)

    def test_request_has_settings(self):
        model = self.app.model
        grid = self.make_grid(key='settings', model_class=model.Setting)

        # paging
        self.assertFalse(grid.request_has_settings('page'))
        with patch.object(grid, 'paginated', new=True):
            with patch.object(self.request, 'GET', new={'pagesize': '20'}):
                self.assertTrue(grid.request_has_settings('page'))
            with patch.object(self.request, 'GET', new={'page': '1'}):
                self.assertTrue(grid.request_has_settings('page'))

        # sorting
        self.assertFalse(grid.request_has_settings('sort'))
        with patch.object(grid, 'sortable', new=True):
            with patch.object(self.request, 'GET', new={'sort1key': 'name'}):
                self.assertTrue(grid.request_has_settings('sort'))

        # filtering
        grid = self.make_grid(key='settings', model_class=model.Setting, filterable=True)
        self.assertFalse(grid.request_has_settings('filter'))
        with patch.object(grid, 'filterable', new=True):
            with patch.object(self.request, 'GET', new={'name': 'john', 'name.verb': 'contains'}):
                self.assertTrue(grid.request_has_settings('filter'))
        with patch.object(self.request, 'GET', new={'filter': '1'}):
            self.assertTrue(grid.request_has_settings('filter'))

    def test_get_setting(self):
        grid = self.make_grid(key='foo')
        settings = {}

        # default is null
        value = grid.get_setting(settings, 'pagesize')
        self.assertIsNone(value)

        # can read value from user session
        self.request.session['grid.foo.pagesize'] = 15
        value = grid.get_setting(settings, 'pagesize', src='session')
        self.assertEqual(value, 15)

        # string value not normalized
        self.request.session['grid.foo.pagesize'] = '15'
        value = grid.get_setting(settings, 'pagesize', src='session')
        self.assertEqual(value, '15')
        self.assertNotEqual(value, 15)

        # but can be normalized
        self.request.session['grid.foo.pagesize'] = '15'
        value = grid.get_setting(settings, 'pagesize', src='session', normalize=int)
        self.assertEqual(value, 15)

        # can read value from request
        self.request.GET = {'pagesize': '25'}
        value = grid.get_setting(settings, 'pagesize', src='request', normalize=int)
        self.assertEqual(value, 25)

        # null when normalization fails
        self.request.GET = {'pagesize': 'invalid'}
        value = grid.get_setting(settings, 'pagesize', src='request', normalize=int)
        self.assertIsNone(value)

        # reset
        del self.request.session['grid.foo.pagesize']
        self.request.GET = {}

        # value can come from provided settings
        settings['pagesize'] = '35'
        value = grid.get_setting(settings, 'pagesize', src='session', normalize=int)
        self.assertEqual(value, 35)

    def test_update_filter_settings(self):
        model = self.app.model

        # nothing happens if not filterable
        grid = self.make_grid(key='settings', model_class=model.Setting)
        settings = {}
        self.request.session['grid.settings.filter.name.active'] = True
        self.request.session['grid.settings.filter.name.verb'] = 'contains'
        self.request.session['grid.settings.filter.name.value'] = 'john'
        grid.update_filter_settings(settings, src='session')
        self.assertEqual(settings, {})

        # nb. now use a filterable grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              filterable=True)

        # settings are updated from session
        settings = {}
        self.request.session['grid.settings.filter.name.active'] = True
        self.request.session['grid.settings.filter.name.verb'] = 'contains'
        self.request.session['grid.settings.filter.name.value'] = 'john'
        grid.update_filter_settings(settings, src='session')
        self.assertTrue(settings['filter.name.active'])
        self.assertEqual(settings['filter.name.verb'], 'contains')
        self.assertEqual(settings['filter.name.value'], 'john')

        # settings are updated from request
        self.request.GET = {'value': 'sally', 'value.verb': 'contains'}
        grid.update_filter_settings(settings, src='request')
        self.assertFalse(settings['filter.name.active'])
        self.assertTrue(settings['filter.value.active'])
        self.assertEqual(settings['filter.value.verb'], 'contains')
        self.assertEqual(settings['filter.value.value'], 'sally')

    def test_update_sort_settings(self):
        model = self.app.model

        # nothing happens if not sortable
        grid = self.make_grid(key='settings', model_class=model.Setting)
        settings = {'sorters.length': 0}
        self.request.session['grid.settings.sorters.length'] = 1
        self.request.session['grid.settings.sorters.1.key'] = 'name'
        self.request.session['grid.settings.sorters.1.dir'] = 'asc'
        grid.update_sort_settings(settings, src='session')
        self.assertEqual(settings['sorters.length'], 0)

        # nb. now use a sortable grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True)

        # settings are updated from session
        settings = {'sorters.length': 1, 'sorters.1.key': 'name', 'sorters.1.dir': 'asc'}
        self.request.session['grid.settings.sorters.length'] = 1
        self.request.session['grid.settings.sorters.1.key'] = 'name'
        self.request.session['grid.settings.sorters.1.dir'] = 'asc'
        grid.update_sort_settings(settings, src='session')
        self.assertEqual(settings['sorters.length'], 1)
        self.assertEqual(settings['sorters.1.key'], 'name')
        self.assertEqual(settings['sorters.1.dir'], 'asc')

        # settings are updated from request
        self.request.GET = {'sort1key': 'value', 'sort1dir': 'desc'}
        grid.update_sort_settings(settings, src='request')
        self.assertEqual(settings['sorters.length'], 1)
        self.assertEqual(settings['sorters.1.key'], 'value')
        self.assertEqual(settings['sorters.1.dir'], 'desc')

    def test_update_page_settings(self):

        # nothing happens if not paginated
        grid = self.make_grid(key='foo')
        settings = {'pagesize': 20, 'page': 1}
        self.request.session['grid.foo.pagesize'] = 10
        self.request.session['grid.foo.page'] = 2
        grid.update_page_settings(settings)
        self.assertEqual(settings['pagesize'], 20)
        self.assertEqual(settings['page'], 1)

        # nb. now use a paginated grid
        grid = self.make_grid(key='foo', paginated=True, paginate_on_backend=True)

        # settings are updated from session
        settings = {'pagesize': 20, 'page': 1}
        self.request.session['grid.foo.pagesize'] = 10
        self.request.session['grid.foo.page'] = 2
        grid.update_page_settings(settings)
        self.assertEqual(settings['pagesize'], 10)
        self.assertEqual(settings['page'], 2)

        # settings are updated from request
        self.request.GET = {'pagesize': '15', 'page': '4'}
        grid.update_page_settings(settings)
        self.assertEqual(settings['pagesize'], 15)
        self.assertEqual(settings['page'], 4)

    def test_persist_settings(self):
        model = self.app.model

        # nb. start out with paginated-only grid
        grid = self.make_grid(key='foo', paginated=True, paginate_on_backend=True)

        # invalid dest
        self.assertRaises(ValueError, grid.persist_settings, {}, dest='doesnotexist')

        # nb. no error if empty settings, but it saves null values
        grid.persist_settings({}, dest='session')
        self.assertIsNone(self.request.session['grid.foo.page'])

        # provided values are saved
        grid.persist_settings({'pagesize': 15, 'page': 3}, dest='session')
        self.assertEqual(self.request.session['grid.foo.page'], 3)

        # nb. now switch to sortable-only grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True)

        # no error if empty settings; does not save values
        grid.persist_settings({}, dest='session')
        self.assertNotIn('grid.settings.sorters.length', self.request.session)

        # provided values are saved
        grid.persist_settings({'sorters.length': 2,
                               'sorters.1.key': 'name',
                               'sorters.1.dir': 'desc',
                               'sorters.2.key': 'value',
                               'sorters.2.dir': 'asc'},
                              dest='session')
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 2)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'desc')
        self.assertEqual(self.request.session['grid.settings.sorters.2.key'], 'value')
        self.assertEqual(self.request.session['grid.settings.sorters.2.dir'], 'asc')

        # old values removed when new are saved
        grid.persist_settings({'sorters.length': 1,
                               'sorters.1.key': 'name',
                               'sorters.1.dir': 'desc'},
                              dest='session')
        self.assertEqual(self.request.session['grid.settings.sorters.length'], 1)
        self.assertEqual(self.request.session['grid.settings.sorters.1.key'], 'name')
        self.assertEqual(self.request.session['grid.settings.sorters.1.dir'], 'desc')
        self.assertNotIn('grid.settings.sorters.2.key', self.request.session)
        self.assertNotIn('grid.settings.sorters.2.dir', self.request.session)

        # nb. now switch to filterable-only grid
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              filterable=True)
        self.assertIn('name', grid.filters)
        self.assertEqual(grid.filters['name'].key, 'name')

        # no error if empty settings; does not save values
        grid.persist_settings({}, dest='session')
        self.assertNotIn('grid.settings.filters.name', self.request.session)

        # provided values are saved
        grid.persist_settings({'filter.name.active': True,
                               'filter.name.verb': 'contains',
                               'filter.name.value': 'john'},
                              dest='session')
        self.assertTrue(self.request.session['grid.settings.filter.name.active'])
        self.assertEqual(self.request.session['grid.settings.filter.name.verb'], 'contains')
        self.assertEqual(self.request.session['grid.settings.filter.name.value'], 'john')

    ##############################
    # sorting methods
    ##############################

    def test_make_backend_sorters(self):
        model = self.app.model

        # default is empty
        grid = self.make_grid()
        sorters = grid.make_backend_sorters()
        self.assertEqual(sorters, {})

        # makes sorters if model class
        grid = self.make_grid(model_class=model.Setting)
        sorters = grid.make_backend_sorters()
        self.assertEqual(len(sorters), 2)
        self.assertIn('name', sorters)
        self.assertIn('value', sorters)

        # does not replace supplied sorters
        grid = self.make_grid(model_class=model.Setting)
        mysorters = {'value': 42}
        sorters = grid.make_backend_sorters(mysorters)
        self.assertEqual(len(sorters), 2)
        self.assertIn('name', sorters)
        self.assertIn('value', sorters)
        self.assertEqual(sorters['value'], 42)
        self.assertEqual(mysorters['value'], 42)

    def test_make_sorter(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'three'},
            {'name': 'foo4', 'value': 'four'},
            {'name': 'foo5', 'value': 'five'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        # plain data
        grid = self.make_grid(columns=['name', 'value'])
        sorter = grid.make_sorter('name')
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(sorted_data[0], {'name': 'foo9', 'value': 'nine'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(sorted_data[0], {'name': 'foo1', 'value': 'ONE'})

        # model class, but still plain data
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('name')
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(sorted_data[0], {'name': 'foo9', 'value': 'nine'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(sorted_data[0], {'name': 'foo1', 'value': 'ONE'})

        # repeat previous test, w/ model property
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter(model.Setting.name)
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(sorted_data[0], {'name': 'foo9', 'value': 'nine'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(sorted_data[0], {'name': 'foo1', 'value': 'ONE'})

        # sqlalchemy query
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('name')
        sorted_query = sorter(sample_query, 'desc')
        sorted_data = sorted_query.all()
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo9', 'value': 'nine'})
        sorted_query = sorter(sample_query, 'asc')
        sorted_data = sorted_query.all()
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo1', 'value': 'ONE'})

        # repeat previous test, w/ model property
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter(model.Setting.name)
        sorted_query = sorter(sample_query, 'desc')
        sorted_data = sorted_query.all()
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo9', 'value': 'nine'})
        sorted_query = sorter(sample_query, 'asc')
        sorted_data = sorted_query.all()
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo1', 'value': 'ONE'})

        # sortfunc for "invalid" column will fail when called; however
        # it can work for manual sort w/ custom keyfunc
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('doesnotexist')
        self.assertRaises(TypeError, sorter, sample_query, 'desc')
        self.assertRaises(KeyError, sorter, sample_data, 'desc')
        sorter = grid.make_sorter('doesnotexist', keyfunc=lambda obj: obj['name'])
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(len(sorted_data), 9)
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(len(sorted_data), 9)

        # case folding is on by default
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('value')
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo2', 'value': 'two'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo8', 'value': 'eight'})

        # results are different with case folding off
        grid = self.make_grid(model_class=model.Setting)
        sorter = grid.make_sorter('value', foldcase=False)
        sorted_data = sorter(sample_data, 'desc')
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo2', 'value': 'two'})
        sorted_data = sorter(sample_data, 'asc')
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo1', 'value': 'ONE'})

    def test_set_joiner(self):

        # basic
        grid = self.make_grid(columns=['foo', 'bar'], sortable=True, sort_on_backend=True)
        self.assertEqual(grid.joiners, {})
        grid.set_joiner('foo', 42)
        self.assertEqual(grid.joiners, {'foo': 42})

    def test_remove_joiner(self):

        # basic
        grid = self.make_grid(columns=['foo', 'bar'], sortable=True, sort_on_backend=True,
                              joiners={'foo': 42})
        self.assertEqual(grid.joiners, {'foo': 42})
        grid.remove_joiner('foo')
        self.assertEqual(grid.joiners, {})

    def test_set_sorter(self):
        model = self.app.model

        # explicit sortfunc
        grid = self.make_grid()
        self.assertEqual(grid.sorters, {})
        sortfunc = lambda data, direction: data
        grid.set_sorter('foo', sortfunc)
        self.assertIs(grid.sorters['foo'], sortfunc)

        # auto from model property
        grid = self.make_grid(model_class=model.Setting, sortable=True, sorters={})
        self.assertEqual(grid.sorters, {})
        grid.set_sorter('name', model.Setting.name)
        self.assertTrue(callable(grid.sorters['name']))

        # auto from column name
        grid = self.make_grid(model_class=model.Setting, sortable=True, sorters={})
        self.assertEqual(grid.sorters, {})
        grid.set_sorter('name', 'name')
        self.assertTrue(callable(grid.sorters['name']))

        # auto from key
        grid = self.make_grid(model_class=model.Setting, sortable=True, sorters={})
        self.assertEqual(grid.sorters, {})
        grid.set_sorter('name')
        self.assertTrue(callable(grid.sorters['name']))

    def test_remove_sorter(self):
        model = self.app.model

        # basics
        grid = self.make_grid(model_class=model.Setting, sortable=True)
        self.assertEqual(len(grid.sorters), 2)
        self.assertIn('name', grid.sorters)
        self.assertIn('value', grid.sorters)
        grid.remove_sorter('value')
        self.assertNotIn('value', grid.sorters)

    def test_set_sort_defaults(self):
        model = self.app.model
        grid = self.make_grid(model_class=model.Setting, sortable=True)
        self.assertEqual(grid.sort_defaults, [])

        # can set just sortkey
        grid.set_sort_defaults('name')
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'asc')])

        # can set sortkey, sortdir
        grid.set_sort_defaults('name', 'desc')
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('name', 'desc')])

        # can set sortkey, sortdir as tuple
        grid.set_sort_defaults(('value', 'asc'))
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('value', 'asc')])

        # can set as list
        grid.sort_multiple = True
        grid.set_sort_defaults([('value', 'asc'), ('name', 'desc')])
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('value', 'asc'),
                                              mod.SortInfo('name', 'desc')])

        # list is pruned if multi-sort disabled
        grid.sort_multiple = False
        grid.set_sort_defaults([('value', 'asc'), ('name', 'desc')])
        self.assertEqual(grid.sort_defaults, [mod.SortInfo('value', 'asc')])

        # error if any other single arg
        self.assertRaises(ValueError, grid.set_sort_defaults, 42)

        # error if more than 2 args
        self.assertRaises(ValueError, grid.set_sort_defaults, 'name', 'asc', 'value', 'desc')

    def test_is_sortable(self):
        model = self.app.model

        # basics, frontend sorting
        grid = self.make_grid(model_class=model.Setting, sortable=True, sort_on_backend=False)
        self.assertTrue(grid.is_sortable('name'))
        self.assertTrue(grid.is_sortable('value'))
        grid.remove_sorter('value')
        # nb. columns are always sortable for frontend, despite remove_sorter()
        self.assertTrue(grid.is_sortable('value'))
        # nb. when grid is not sortable, no column is either
        grid.sortable = False
        self.assertFalse(grid.is_sortable('name'))

        # same test but with backend sorting
        grid = self.make_grid(model_class=model.Setting, sortable=True, sort_on_backend=True)
        self.assertTrue(grid.is_sortable('name'))
        self.assertTrue(grid.is_sortable('value'))
        grid.remove_sorter('value')
        self.assertFalse(grid.is_sortable('value'))
        # nb. when grid is not sortable, no column is either
        grid.sortable = False
        self.assertFalse(grid.is_sortable('name'))

    def test_make_backend_filters(self):
        model = self.app.model

        # default is empty
        grid = self.make_grid()
        filters = grid.make_backend_filters()
        self.assertEqual(filters, {})

        # makes filters if model class
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.Setting)
            filters = grid.make_backend_filters()
        self.assertEqual(len(filters), 2)
        self.assertIn('name', filters)
        self.assertIn('value', filters)

        # does not replace supplied filters
        myfilters = {'value': 42}
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.Setting)
            filters = grid.make_backend_filters(myfilters)
        self.assertEqual(len(filters), 2)
        self.assertIn('name', filters)
        self.assertIn('value', filters)
        self.assertEqual(filters['value'], 42)
        self.assertEqual(myfilters['value'], 42)

        # filters for all *true* columns by default, despite grid.columns
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.User,
                                  columns=['username', 'person'])
            filters = grid.make_backend_filters()
            self.assertIn('username', filters)
            self.assertIn('active', filters)
            # nb. relationship not included by default
            self.assertNotIn('person', filters)

    def test_make_filter(self):
        model = self.app.model

        # arg is column name
        grid = self.make_grid(model_class=model.Setting)
        filtr = grid.make_filter('name')
        self.assertIsInstance(filtr, StringAlchemyFilter)

        # arg is column name, but model class is invalid
        grid = self.make_grid(model_class=42)
        self.assertRaises(ValueError, grid.make_filter, 'name')

        # arg is model property
        grid = self.make_grid(model_class=model.Setting)
        filtr = grid.make_filter(model.Setting.name)
        self.assertIsInstance(filtr, StringAlchemyFilter)

        # model property as kwarg
        grid = self.make_grid(model_class=model.Setting)
        filtr = grid.make_filter(None, model_property=model.Setting.name)
        self.assertIsInstance(filtr, StringAlchemyFilter)

        # default factory
        grid = self.make_grid(model_class=model.Setting)
        with patch.dict(default_sqlalchemy_filters, {None: GridFilter}, clear=True):
            filtr = grid.make_filter(model.Setting.name)
        self.assertIsInstance(filtr, GridFilter)
        self.assertNotIsInstance(filtr, StringAlchemyFilter)

        # factory override
        grid = self.make_grid(model_class=model.Setting)
        filtr = grid.make_filter(model.Setting.name, factory=GridFilter)
        self.assertIsInstance(filtr, GridFilter)
        self.assertNotIsInstance(filtr, StringAlchemyFilter)

    def test_set_filter(self):
        model = self.app.model

        with patch.object(mod.Grid, 'make_filter', return_value=42):

            # auto from model property
            grid = self.make_grid(model_class=model.Setting)
            self.assertEqual(grid.filters, {})
            grid.set_filter('name', model.Setting.name)
            self.assertIn('name', grid.filters)

            # auto from column name
            grid = self.make_grid(model_class=model.Setting)
            self.assertEqual(grid.filters, {})
            grid.set_filter('name', 'name')
            self.assertIn('name', grid.filters)

            # auto from key
            grid = self.make_grid(model_class=model.Setting)
            self.assertEqual(grid.filters, {})
            grid.set_filter('name')
            self.assertIn('name', grid.filters)

            # explicit is not yet implemented
            grid = self.make_grid(model_class=model.Setting)
            self.assertEqual(grid.filters, {})
            self.assertRaises(NotImplementedError, grid.set_filter, 'name', lambda q: q)

    def test_remove_filter(self):
        model = self.app.model

        # basics
        with patch.object(mod.Grid, 'make_filter'):
            # nb. filters are MagicMock instances
            grid = self.make_grid(model_class=model.Setting, filterable=True)
        self.assertEqual(len(grid.filters), 2)
        self.assertIn('name', grid.filters)
        self.assertIn('value', grid.filters)
        grid.remove_filter('value')
        self.assertNotIn('value', grid.filters)

    def test_set_filter_defaults(self):
        model = self.app.model

        # empty by default
        grid = self.make_grid(model_class=model.Setting, filterable=True)
        self.assertEqual(grid.filter_defaults, {})

        # can specify via method call
        grid.set_filter_defaults(name={'active': True})
        self.assertEqual(grid.filter_defaults, {'name': {'active': True}})

        # can specify via constructor
        grid = self.make_grid(model_class=model.Setting, filterable=True,
                              filter_defaults={'name': {'active': True}})
        self.assertEqual(grid.filter_defaults, {'name': {'active': True}})

    ##############################
    # data methods
    ##############################

    def test_get_visible_data(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'three'},
            {'name': 'foo4', 'value': 'four'},
            {'name': 'foo5', 'value': 'five'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        # data is sorted and paginated
        grid = self.make_grid(model_class=model.Setting,
                              data=sample_query,
                              filterable=True,
                              sortable=True, sort_on_backend=True,
                              sort_defaults=('name', 'desc'),
                              paginated=True, paginate_on_backend=True,
                              pagesize=4, page=2)
        grid.load_settings()
        # nb. for now the filtering is mocked
        with patch.object(grid, 'filter_data') as filter_data:
            filter_data.side_effect = lambda q: q
            visible = grid.get_visible_data()
            filter_data.assert_called_once_with(sample_query)
        self.assertEqual([s.name for s in visible], ['foo5', 'foo4', 'foo3', 'foo2'])

    def test_filter_data(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        grid = self.make_grid(key='settings', model_class=model.Setting, filterable=True)
        self.assertEqual(list(grid.filters), ['name', 'value'])
        self.assertIsInstance(grid.filters['name'], StringAlchemyFilter)
        self.assertIsInstance(grid.filters['value'], StringAlchemyFilter)

        # not filtered by default
        grid.load_settings()
        self.assertEqual(grid.active_filters, [])
        filtered_query = grid.filter_data(sample_query)
        self.assertIs(filtered_query, sample_query)

        # can be filtered per session settings
        self.request.session['grid.settings.filter.value.active'] = True
        self.request.session['grid.settings.filter.value.verb'] = 'contains'
        self.request.session['grid.settings.filter.value.value'] = 'ggg'
        grid.load_settings()
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].key, 'value')
        filtered_query = grid.filter_data(sample_query)
        self.assertIsInstance(filtered_query, orm.Query)
        self.assertIsNot(filtered_query, sample_query)
        self.assertEqual(filtered_query.count(), 3)

        # can be filtered per request settings
        self.request.GET = {'value': 's', 'value.verb': 'contains'}
        grid.load_settings()
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].key, 'value')
        filtered_query = grid.filter_data(sample_query)
        self.assertIsInstance(filtered_query, orm.Query)
        self.assertEqual(filtered_query.count(), 2)

        # not filtered if verb is invalid
        self.request.GET = {'value': 'ggg', 'value.verb': 'doesnotexist'}
        grid.load_settings()
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].verb, 'doesnotexist')
        filtered_query = grid.filter_data(sample_query)
        self.assertIs(filtered_query, sample_query)
        self.assertEqual(filtered_query.count(), 9)

        # not filtered if error
        self.request.GET = {'value': 'ggg', 'value.verb': 'contains'}
        grid.load_settings()
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].verb, 'contains')
        filtered_query = grid.filter_data(sample_query)
        self.assertIsNot(filtered_query, sample_query)
        self.assertEqual(filtered_query.count(), 3)
        with patch.object(grid.active_filters[0], 'filter_contains', side_effect=RuntimeError):
            filtered_query = grid.filter_data(sample_query)
        self.assertIs(filtered_query, sample_query)
        self.assertEqual(filtered_query.count(), 9)

        # joiner is invoked
        self.assertEqual(len(grid.active_filters), 1)
        self.assertEqual(grid.active_filters[0].key, 'value')
        joiner = MagicMock(side_effect=lambda q: q)
        grid.joiners = {'value': joiner}
        grid.joined = set()
        filtered_query = grid.filter_data(sample_query)
        joiner.assert_called_once_with(sample_query)
        self.assertEqual(filtered_query.count(), 3)

    def test_sort_data(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'ggg'},
            {'name': 'foo4', 'value': 'ggg'},
            {'name': 'foo5', 'value': 'ggg'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        grid = self.make_grid(model_class=model.Setting,
                              sortable=True, sort_on_backend=True,
                              sort_defaults=('name', 'desc'))
        grid.load_settings()

        # can sort a simple list of data
        sorted_data = grid.sort_data(sample_data)
        self.assertIsInstance(sorted_data, list)
        self.assertEqual(len(sorted_data), 9)
        self.assertEqual(sorted_data[0]['name'], 'foo9')
        self.assertEqual(sorted_data[-1]['name'], 'foo1')

        # can also sort a data query
        sorted_query = grid.sort_data(sample_query)
        self.assertIsInstance(sorted_query, orm.Query)
        sorted_data = sorted_query.all()
        self.assertEqual(len(sorted_data), 9)
        self.assertEqual(sorted_data[0]['name'], 'foo9')
        self.assertEqual(sorted_data[-1]['name'], 'foo1')

        # cannot sort data if sorter missing in overrides
        sorted_data = grid.sort_data(sample_data, sorters=[])
        # nb. sorted data is in same order as original sample (not sorted)
        self.assertEqual(sorted_data[0]['name'], 'foo1')
        self.assertEqual(sorted_data[-1]['name'], 'foo9')

        # multi-column sorting for list data
        sorted_data = grid.sort_data(sample_data, sorters=[{'key': 'value', 'dir': 'asc'},
                                                           {'key': 'name', 'dir': 'asc'}])
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo8', 'value': 'eight'})
        self.assertEqual(dict(sorted_data[1]), {'name': 'foo3', 'value': 'ggg'})
        self.assertEqual(dict(sorted_data[3]), {'name': 'foo5', 'value': 'ggg'})
        self.assertEqual(dict(sorted_data[-1]), {'name': 'foo2', 'value': 'two'})

        # multi-column sorting for query
        sorted_query = grid.sort_data(sample_query, sorters=[{'key': 'value', 'dir': 'asc'},
                                                             {'key': 'name', 'dir': 'asc'}])
        self.assertEqual(dict(sorted_data[0]), {'name': 'foo8', 'value': 'eight'})
        self.assertEqual(dict(sorted_data[1]), {'name': 'foo3', 'value': 'ggg'})
        self.assertEqual(dict(sorted_data[3]), {'name': 'foo5', 'value': 'ggg'})
        self.assertEqual(dict(sorted_data[-1]), {'name': 'foo2', 'value': 'two'})

        # cannot sort data if sortfunc is missing for column
        grid.remove_sorter('name')
        sorted_data = grid.sort_data(sample_data, sorters=[{'key': 'value', 'dir': 'asc'},
                                                           {'key': 'name', 'dir': 'asc'}])
        # nb. sorted data is in same order as original sample (not sorted)
        self.assertEqual(sorted_data[0]['name'], 'foo1')
        self.assertEqual(sorted_data[-1]['name'], 'foo9')

        # now try with a joiner
        query = self.session.query(model.User)
        grid = self.make_grid(model_class=model.User,
                              data=query,
                              columns=['username', 'full_name'],
                              sortable=True, sort_on_backend=True,
                              sort_defaults='full_name',
                              joiners={
                                  'full_name': lambda q: q.join(model.Person),
                              })
        grid.set_sorter('full_name', model.Person.full_name)
        grid.load_settings()
        data = grid.get_visible_data()
        self.assertIsInstance(data, orm.Query)

    def test_paginate_data(self):
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'three'},
            {'name': 'foo4', 'value': 'four'},
            {'name': 'foo5', 'value': 'five'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        # basic list pager
        grid = self.make_grid(paginated=True, paginate_on_backend=True)
        pager = grid.paginate_data(sample_data)
        self.assertIsInstance(pager, Page)

        # basic query pager
        grid = self.make_grid(paginated=True, paginate_on_backend=True)
        pager = grid.paginate_data(sample_query)
        self.assertIsInstance(pager, SqlalchemyOrmPage)

        # page is reset to 1 for empty data
        self.request.session['grid.foo.page'] = 2
        grid = self.make_grid(key='foo', paginated=True, paginate_on_backend=True)
        grid.load_settings()
        self.assertEqual(grid.page, 2)
        self.assertEqual(self.request.session['grid.foo.page'], 2)
        pager = grid.paginate_data(sample_data)
        self.assertEqual(pager.page, 1)
        self.assertEqual(grid.page, 1)
        self.assertEqual(self.request.session['grid.foo.page'], 1)

    ##############################
    # rendering methods
    ##############################

    def test_render_batch_id(self):
        grid = self.make_grid(columns=['foo', 'bar'])

        # null
        obj = MagicMock(foo=None)
        self.assertEqual(grid.render_batch_id(obj, 'foo', None), "")

        # int
        obj = MagicMock(foo=42)
        self.assertEqual(grid.render_batch_id(obj, 'foo', 42), "00000042")

    def test_render_boolean(self):
        grid = self.make_grid(columns=['foo', 'bar'])

        # null
        obj = MagicMock(foo=None)
        self.assertEqual(grid.render_boolean(obj, 'foo', None), "")

        # true
        obj = MagicMock(foo=True)
        self.assertEqual(grid.render_boolean(obj, 'foo', True), "Yes")

        # false
        obj = MagicMock(foo=False)
        self.assertEqual(grid.render_boolean(obj, 'foo', False), "No")

    def test_render_currency(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        obj = MagicMock()

        # null
        self.assertEqual(grid.render_currency(obj, 'foo', None), '')

        # basic decimal example
        value = decimal.Decimal('42.00')
        self.assertEqual(grid.render_currency(obj, 'foo', value), '$42.00')

        # basic float example
        value = 42.00
        self.assertEqual(grid.render_currency(obj, 'foo', value), '$42.00')

        # decimal places will be rounded
        value = decimal.Decimal('42.12345')
        self.assertEqual(grid.render_currency(obj, 'foo', value), '$42.12')

        # negative numbers get parens
        value = decimal.Decimal('-42.42')
        self.assertEqual(grid.render_currency(obj, 'foo', value), '($42.42)')

    def test_render_quantity(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        obj = MagicMock()

        # null
        self.assertEqual(grid.render_quantity(obj, 'foo', None), "")

        # integer decimals become integers
        value = decimal.Decimal('1.000')
        self.assertEqual(grid.render_quantity(obj, 'foo', value), "1")

        # but decimal places are preserved
        value = decimal.Decimal('1.234')
        self.assertEqual(grid.render_quantity(obj ,'foo', value), "1.234")

        # zero is *not* empty string (with this renderer)
        self.assertEqual(grid.render_quantity(obj, 'foo', 0), "0")

    def test_render_date(self):
        grid = self.make_grid(columns=['foo', 'bar'])

        # null
        obj = MagicMock(dt=None)
        result = grid.render_date(obj, 'dt', None)
        self.assertIsNone(result)

        # typical
        dt = datetime.date(2025, 1, 13)
        obj = MagicMock(dt=dt)
        result = grid.render_date(obj, 'dt', str(dt))
        self.assertEqual(result, '2025-01-13')

    def test_render_datetime(self):
        grid = self.make_grid(columns=['foo', 'bar'])

        obj = MagicMock(dt=None)
        result = grid.render_datetime(obj, 'dt', None)
        self.assertIsNone(result)

        dt = datetime.datetime(2024, 12, 12, 13, 44, tzinfo=datetime.timezone.utc)
        obj = MagicMock(dt=dt)
        result = grid.render_datetime(obj, 'dt', str(dt))
        self.assertEqual(result, '2024-12-12 13:44+0000')

    def test_render_vue_tag(self):
        grid = self.make_grid(columns=['foo', 'bar'])
        html = grid.render_vue_tag()
        self.assertEqual(html, '<wutta-grid></wutta-grid>')

    def test_render_vue_template(self):
        self.pyramid_config.include('pyramid_mako')
        self.pyramid_config.add_subscriber('wuttaweb.subscribers.before_render',
                                           'pyramid.events.BeforeRender')

        grid = self.make_grid(columns=['foo', 'bar'])
        html = grid.render_vue_template()
        self.assertIn('<script type="text/x-template" id="wutta-grid-template">', html)

    def test_render_table_element(self):
        self.pyramid_config.include('pyramid_mako')
        self.pyramid_config.add_subscriber('wuttaweb.subscribers.before_render',
                                           'pyramid.events.BeforeRender')

        grid = self.make_grid(key='foobar', columns=['foo', 'bar'])

        # form not required
        html = grid.render_table_element()
        self.assertNotIn('<script ', html)
        self.assertIn('<b-table ', html)

        # form will register grid data
        form = Form(self.request)
        self.assertEqual(len(form.grid_vue_context), 0)
        html = grid.render_table_element(form)
        self.assertEqual(len(form.grid_vue_context), 1)
        self.assertIn('foobar', form.grid_vue_context)

    def test_render_vue_finalize(self):
        grid = self.make_grid()
        html = grid.render_vue_finalize()
        self.assertIn('<script>', html)
        self.assertIn("Vue.component('wutta-grid', WuttaGrid)", html)

    def test_get_vue_columns(self):

        # error if no columns are set
        grid = self.make_grid()
        self.assertRaises(ValueError, grid.get_vue_columns)

        # otherwise get back field/label dicts
        grid = self.make_grid(columns=['foo', 'bar'])
        columns = grid.get_vue_columns()
        first = columns[0]
        self.assertEqual(first['field'], 'foo')
        self.assertEqual(first['label'], 'Foo')

    def test_get_vue_active_sorters(self):
        model = self.app.model

        # empty
        grid = self.make_grid(key='foo', sortable=True, sort_on_backend=True)
        grid.load_settings()
        sorters = grid.get_vue_active_sorters()
        self.assertEqual(sorters, [])

        # format is different
        grid = self.make_grid(key='settings', model_class=model.Setting,
                              sortable=True, sort_on_backend=True,
                              sort_defaults='name')
        grid.load_settings()
        self.assertEqual(grid.active_sorters, [{'key': 'name', 'dir': 'asc'}])
        sorters = grid.get_vue_active_sorters()
        self.assertEqual(sorters, [{'field': 'name', 'order': 'asc'}])

    def test_get_vue_filters(self):
        model = self.app.model

        # basic
        grid = self.make_grid(key='settings', model_class=model.Setting, filterable=True)
        grid.load_settings()
        filters = grid.get_vue_filters()
        self.assertEqual(len(filters), 2)

    def test_object_to_dict(self):
        grid = self.make_grid()
        setting = {'name': 'foo', 'value': 'bar'}

        # new dict but with same values
        dct = grid.object_to_dict(setting)
        self.assertIsInstance(dct, dict)
        self.assertIsNot(dct, setting)
        self.assertEqual(dct, setting)

        # random object, not iterable
        class MockSetting:
            def __init__(self, **kw):
                self.__dict__.update(kw)
        mock = MockSetting(**setting)
        dct = grid.object_to_dict(mock)
        self.assertIsInstance(dct, dict)
        self.assertEqual(dct, setting)

    def test_get_vue_context(self):

        # empty if no columns defined
        grid = self.make_grid()
        context = grid.get_vue_context()
        self.assertEqual(context, {'data': [], 'row_classes': {}})

        # typical data is a list
        mydata = [
            {'foo': 'bar'},
        ]
        grid = self.make_grid(columns=['foo'], data=mydata)
        context = grid.get_vue_context()
        self.assertEqual(context, {'data': [{'foo': 'bar'}], 'row_classes': {}})

        # if grid has actions, that list may be supplemented
        grid.actions.append(mod.GridAction(self.request, 'view', url='/blarg'))
        context = grid.get_vue_context()
        self.assertIsNot(context['data'], mydata)
        self.assertEqual(context, {'data': [{'foo': 'bar', '_action_url_view': '/blarg'}],
                                   'row_classes': {}})

        # can override value rendering
        grid.set_renderer('foo', lambda record, key, value: "blah blah")
        context = grid.get_vue_context()
        self.assertEqual(context, {'data': [{'foo': 'blah blah', '_action_url_view': '/blarg'}],
                                   'row_classes': {}})

        # can set row class
        grid.row_class = 'whatever'
        context = grid.get_vue_context()
        self.assertEqual(context, {'data': [{'foo': 'blah blah', '_action_url_view': '/blarg'}],
                                   'row_classes': {'0': 'whatever'}})

    def test_get_vue_data(self):

        # empty if no columns defined
        grid = self.make_grid()
        data = grid.get_vue_data()
        self.assertEqual(data, [])

        # typical data is a list
        mydata = [
            {'foo': 'bar'},
        ]
        grid = self.make_grid(columns=['foo'], data=mydata)
        data = grid.get_vue_data()
        self.assertEqual(data, [{'foo': 'bar'}])

        # if grid has actions, that list may be supplemented
        grid.actions.append(mod.GridAction(self.request, 'view', url='/blarg'))
        data = grid.get_vue_data()
        self.assertIsNot(data, mydata)
        self.assertEqual(data, [{'foo': 'bar', '_action_url_view': '/blarg'}])

        # can override value rendering
        grid.set_renderer('foo', lambda record, key, value: "blah blah")
        data = grid.get_vue_data()
        self.assertEqual(data, [{'foo': 'blah blah', '_action_url_view': '/blarg'}])

    def test_get_row_class(self):
        model = self.app.model
        user = model.User(username='barney', active=True)
        self.session.add(user)
        self.session.commit()
        data = dict(user)

        # null by default
        grid = self.make_grid()
        self.assertIsNone(grid.get_row_class(user, data, 1))

        # can use static class
        grid.row_class = 'foo'
        self.assertEqual(grid.get_row_class(user, data, 1), 'foo')

        # can use callable
        def status(u, d, i):
            if not u.active:
                return 'inactive'
        grid.row_class = status
        self.assertIsNone(grid.get_row_class(user, data, 1))
        user.active = False
        self.assertEqual(grid.get_row_class(user, data, 1), 'inactive')

    def test_get_vue_pager_stats(self):
        data = [
            {'foo': 1, 'bar': 1},
            {'foo': 2, 'bar': 2},
            {'foo': 3, 'bar': 3},
            {'foo': 4, 'bar': 4},
            {'foo': 5, 'bar': 5},
            {'foo': 6, 'bar': 6},
            {'foo': 7, 'bar': 7},
            {'foo': 8, 'bar': 8},
            {'foo': 9, 'bar': 9},
        ]

        grid = self.make_grid(columns=['foo', 'bar'], pagesize=4, page=2)
        grid.pager = grid.paginate_data(data)
        stats = grid.get_vue_pager_stats()
        self.assertEqual(stats['item_count'], 9)
        self.assertEqual(stats['items_per_page'], 4)
        self.assertEqual(stats['page'], 2)
        self.assertEqual(stats['first_item'], 5)
        self.assertEqual(stats['last_item'], 8)


class TestGridAction(TestCase):

    def setUp(self):
        self.config = WuttaConfig()
        self.request = testing.DummyRequest(wutta_config=self.config, use_oruga=False)

    def make_action(self, key, **kwargs):
        return mod.GridAction(self.request, key, **kwargs)

    def test_render_icon(self):

        # icon is derived from key by default
        action = self.make_action('blarg')
        html = action.render_icon()
        self.assertIn('<i class="fas fa-blarg">', html)

        # oruga has different output
        self.request.use_oruga = True
        html = action.render_icon()
        self.assertIn('<o-icon icon="blarg">', html)

    def test_render_label(self):

        # label is derived from key by default
        action = self.make_action('blarg')
        label = action.render_label()
        self.assertEqual(label, "Blarg")

        # otherwise use what caller provides
        action = self.make_action('foo', label="Bar")
        label = action.render_label()
        self.assertEqual(label, "Bar")

    def test_render_icon_and_label(self):
        action = self.make_action('blarg')
        with patch.multiple(action,
                            render_icon=lambda: 'ICON',
                            render_label=lambda: 'LABEL'):
            html = action.render_icon_and_label()
            self.assertEqual('ICON LABEL', html)

    def test_get_url(self):
        obj = {'foo': 'bar'}

        # null by default
        action = self.make_action('blarg')
        url = action.get_url(obj)
        self.assertIsNone(url)

        # or can be "static"
        action = self.make_action('blarg', url='/foo')
        url = action.get_url(obj)
        self.assertEqual(url, '/foo')

        # or can be "dynamic"
        action = self.make_action('blarg', url=lambda o, i: '/yeehaw')
        url = action.get_url(obj)
        self.assertEqual(url, '/yeehaw')



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from importlib.metadata import version as get_version

project = 'WuttaWeb'
copyright = '2024, Lance Edgar'
author = 'Lance Edgar'
release = get_version('WuttaWeb')

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.todo',
    'sphinxcontrib.programoutput',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

intersphinx_mapping = {
    'colander': ('https://docs.pylonsproject.org/projects/colander/en/latest/', None),
    'deform': ('https://docs.pylonsproject.org/projects/deform/en/latest/', None),
    'fanstatic': ('https://www.fanstatic.org/en/latest/', None),
    'pyramid': ('https://docs.pylonsproject.org/projects/pyramid/en/latest/', None),
    'python': ('https://docs.python.org/3/', None),
    'rattail-manual': ('https://rattailproject.org/docs/rattail-manual/', None),
    'sqlalchemy': ('http://docs.sqlalchemy.org/en/latest/', None),
    'webhelpers2': ('https://webhelpers2.readthedocs.io/en/latest/', None),
    'wuttjamaican': ('https://rattailproject.org/docs/wuttjamaican/', None),
    'wutta-continuum': ('https://rattailproject.org/docs/wutta-continuum/', None),
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_static_path = ['_static']



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Form schema types
"""

import datetime
import uuid as _uuid

import colander
import sqlalchemy as sa

from wuttjamaican.db.model import Person
from wuttjamaican.conf import parse_list

from wuttaweb.db import Session
from wuttaweb.forms import widgets


class WuttaDateTime(colander.DateTime):
    """
    Custom schema type for ``datetime`` fields.

    This should be used automatically for
    :class:`sqlalchemy:sqlalchemy.types.DateTime` columns unless you
    register another default.

    This schema type exists for sake of convenience, when working with
    the Buefy datepicker + timepicker widgets.
    """

    def deserialize(self, node, cstruct):
        """ """
        if not cstruct:
            return colander.null

        formats = [
            '%Y-%m-%dT%H:%M:%S',
            '%Y-%m-%dT%I:%M %p',
        ]

        for fmt in formats:
            try:
                return datetime.datetime.strptime(cstruct, fmt)
            except:
                pass

        node.raise_invalid("Invalid date and/or time")


class ObjectNode(colander.SchemaNode):
    """
    Custom schema node class which adds methods for compatibility with
    ColanderAlchemy.  This is a direct subclass of
    :class:`colander:colander.SchemaNode`.

    ColanderAlchemy will call certain methods on any node found in the
    schema.  However these methods are not "standard" and only exist
    for ColanderAlchemy nodes.

    So we must add nodes using this class, to ensure the node has all
    methods needed by ColanderAlchemy.
    """

    def dictify(self, obj):
        """
        This method is called by ColanderAlchemy when translating the
        in-app Python object to a value suitable for use in the form
        data dict.

        The logic here will look for a ``dictify()`` method on the
        node's "type" instance (``self.typ``; see also
        :class:`colander:colander.SchemaNode`) and invoke it if found.

        For an example type which is supported in this way, see
        :class:`ObjectRef`.

        If the node's type does not have a ``dictify()`` method, this
        will just convert the object to a string and return that.
        """
        if hasattr(self.typ, 'dictify'):
            return self.typ.dictify(obj)

        # TODO: this is better than raising an error, as it previously
        # did, but seems like troubleshooting problems may often lead
        # one here.. i suspect this needs to do something smarter but
        # not sure what that is yet
        return str(obj)

    def objectify(self, value):
        """
        This method is called by ColanderAlchemy when translating form
        data to the final Python representation.

        The logic here will look for an ``objectify()`` method on the
        node's "type" instance (``self.typ``; see also
        :class:`colander:colander.SchemaNode`) and invoke it if found.

        For an example type which is supported in this way, see
        :class:`ObjectRef`.

        If the node's type does not have an ``objectify()`` method,
        this will raise ``NotImplementeError``.
        """
        if hasattr(self.typ, 'objectify'):
            return self.typ.objectify(value)

        class_name = self.typ.__class__.__name__
        raise NotImplementedError(f"you must define {class_name}.objectify()")


class WuttaEnum(colander.Enum):
    """
    Custom schema type for enum fields.

    This is a subclass of :class:`colander.Enum`, but adds a
    default widget (``SelectWidget``) with enum choices.

    :param request: Current :term:`request` object.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def widget_maker(self, **kwargs):
        """ """

        if 'values' not in kwargs:
            kwargs['values'] = [(getattr(e, self.attr), getattr(e, self.attr))
                                for e in self.enum_cls]

        return widgets.SelectWidget(**kwargs)


class WuttaDictEnum(colander.String):
    """
    Schema type for "pseudo-enum" fields which reference a dict for
    known values instead of a true enum class.

    This is primarily for use with "status" fields such as
    :attr:`~wuttjamaican:wuttjamaican.db.model.batch.BatchRowMixin.status_code`.

    This is a subclass of :class:`colander.String`, but adds a default
    widget (``SelectWidget``) with enum choices.

    :param request: Current :term:`request` object.

    :param enum_dct: Dict with possible enum values and labels.
    """

    def __init__(self, request, enum_dct, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()
        self.enum_dct = enum_dct

    def widget_maker(self, **kwargs):
        """ """
        if 'values' not in kwargs:
            kwargs['values'] = [(k, v) for k, v in self.enum_dct.items()]

        return widgets.SelectWidget(**kwargs)


class WuttaMoney(colander.Money):
    """
    Custom schema type for "money" fields.

    This is a subclass of :class:`colander:colander.Money`, but uses
    the custom :class:`~wuttaweb.forms.widgets.WuttaMoneyInputWidget`
    by default.

    :param request: Current :term:`request` object.

    :param scale: If this kwarg is specified, it will be passed along
       to the widget constructor.
    """

    def __init__(self, request, *args, **kwargs):
        self.scale = kwargs.pop('scale', None)
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def widget_maker(self, **kwargs):
        """ """
        if self.scale:
            kwargs.setdefault('scale', self.scale)
        return widgets.WuttaMoneyInputWidget(self.request, **kwargs)


class WuttaQuantity(colander.Decimal):
    """
    Custom schema type for "quantity" fields.

    This is a subclass of :class:`colander:colander.Decimal` but will
    serialize values via
    :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_quantity()`.

    :param request: Current :term:`request` object.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, node, appstruct):
        """ """
        if appstruct in (colander.null, None):
            return colander.null

        # nb. we render as quantity here to avoid values like 12.0000,
        # so we just show value like 12 instead
        return self.app.render_quantity(appstruct)


class WuttaSet(colander.Set):
    """
    Custom schema type for :class:`python:set` fields.

    This is a subclass of :class:`colander.Set`.

    :param request: Current :term:`request` object.
    """

    def __init__(self, request):
        super().__init__()
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()


class ObjectRef(colander.SchemaType):
    """
    Custom schema type for a model class reference field.

    This expects the incoming ``appstruct`` to be either a model
    record instance, or ``None``.

    Serializes to the instance UUID as string, or ``colander.null``;
    form data should be of the same nature.

    This schema type is not useful directly, but various other types
    will subclass it.  Each should define (at least) the
    :attr:`model_class` attribute or property.

    :param request: Current :term:`request` object.

    :param empty_option: If a select widget is used, this determines
       whether an empty option is included for the dropdown.  Set
       this to one of the following to add an empty option:

       * ``True`` to add the default empty option
       * label text for the empty option
       * tuple of ``(value, label)`` for the empty option

       Note that in the latter, ``value`` must be a string.
    """

    default_empty_option = ('', "(none)")

    def __init__(
            self,
            request,
            empty_option=None,
            *args,
            **kwargs,
    ):
        # nb. allow session injection for tests
        self.session = kwargs.pop('session', Session())
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()
        self.model_instance = None

        if empty_option:
            if empty_option is True:
                self.empty_option = self.default_empty_option
            elif isinstance(empty_option, tuple) and len(empty_option) == 2:
                self.empty_option = empty_option
            else:
                self.empty_option = ('', str(empty_option))
        else:
            self.empty_option = None

    @property
    def model_class(self):
        """
        Should be a reference to the model class to which this schema
        type applies
        (e.g. :class:`~wuttjamaican:wuttjamaican.db.model.base.Person`).
        """
        class_name = self.__class__.__name__
        raise NotImplementedError(f"you must define {class_name}.model_class")

    def serialize(self, node, appstruct):
        """ """
        # nb. normalize to empty option if no object ref, so that
        # works as expected
        if self.empty_option and not appstruct:
            return self.empty_option[0]

        if appstruct is colander.null:
            return colander.null

        # nb. keep a ref to this for later use
        node.model_instance = appstruct

        # serialize to PK as string
        return self.serialize_object(appstruct)

    def serialize_object(self, obj):
        """
        Serialize the given object to its primary key as string.

        Default logic assumes the object has a UUID; subclass can
        override as needed.

        :param obj: Object reference for the node.

        :returns: Object primary key as string.
        """
        return obj.uuid.hex

    def deserialize(self, node, cstruct):
        """ """
        if not cstruct:
            return colander.null

        # nb. use shortcut to fetch model instance from DB
        return self.objectify(cstruct)

    def dictify(self, obj):
        """ """

        # TODO: would we ever need to do something else?
        return obj

    def objectify(self, value):
        """
        For the given UUID value, returns the object it represents
        (based on :attr:`model_class`).

        If the value is empty, returns ``None``.

        If the value is not empty but object cannot be found, raises
        ``colander.Invalid``.
        """
        if not value:
            return

        if isinstance(value, self.model_class):
            return value

        # fetch object from DB
        model = self.app.model
        obj = None
        if isinstance(value, _uuid.UUID):
            obj = self.session.get(self.model_class, value)
        else:
            try:
                obj = self.session.get(self.model_class, _uuid.UUID(value))
            except ValueError:
                pass

        # raise error if not found
        if not obj:
            class_name = self.model_class.__name__
            raise ValueError(f"{class_name} not found: {value}")

        return obj

    def get_query(self):
        """
        Returns the main SQLAlchemy query responsible for locating the
        dropdown choices for the select widget.

        This is called by :meth:`widget_maker()`.
        """
        query = self.session.query(self.model_class)
        query = self.sort_query(query)
        return query

    def sort_query(self, query):
        """
        TODO
        """
        return query

    def widget_maker(self, **kwargs):
        """
        This method is responsible for producing the default widget
        for the schema node.

        Deform calls this method automatically when constructing the
        default widget for a field.

        :returns: Instance of
           :class:`~wuttaweb.forms.widgets.ObjectRefWidget`.
        """

        if 'values' not in kwargs:
            query = self.get_query()
            objects = query.all()
            values = [(self.serialize_object(obj), str(obj))
                      for obj in objects]
            if self.empty_option:
                values.insert(0, self.empty_option)
            kwargs['values'] = values

        if 'url' not in kwargs:
            kwargs['url'] = self.get_object_url

        return widgets.ObjectRefWidget(self.request, **kwargs)

    def get_object_url(self, obj):
        """
        Returns the "view" URL for the given object, if applicable.

        This is used when rendering the field readonly.  If this
        method returns a URL then the field text will be wrapped with
        a hyperlink, otherwise it will be shown as-is.

        Default logic always returns ``None``; subclass should
        override as needed.
        """


class PersonRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttjamaican:wuttjamaican.db.model.base.Person` reference
    field.

    This is a subclass of :class:`ObjectRef`.
    """

    @property
    def model_class(self):
        """ """
        model = self.app.model
        return model.Person

    def sort_query(self, query):
        """ """
        return query.order_by(self.model_class.full_name)

    def get_object_url(self, person):
        """ """
        return self.request.route_url('people.view', uuid=person.uuid)


class RoleRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttjamaican:wuttjamaican.db.model.auth.Role` reference
    field.

    This is a subclass of :class:`ObjectRef`.
    """

    @property
    def model_class(self):
        """ """
        model = self.app.model
        return model.Role

    def sort_query(self, query):
        """ """
        return query.order_by(self.model_class.name)

    def get_object_url(self, role):
        """ """
        return self.request.route_url('roles.view', uuid=role.uuid)


class UserRef(ObjectRef):
    """
    Custom schema type for a
    :class:`~wuttjamaican:wuttjamaican.db.model.auth.User` reference
    field.

    This is a subclass of :class:`ObjectRef`.
    """

    @property
    def model_class(self):
        """ """
        model = self.app.model
        return model.User

    def sort_query(self, query):
        """ """
        return query.order_by(self.model_class.username)

    def get_object_url(self, user):
        """ """
        return self.request.route_url('users.view', uuid=user.uuid)


class RoleRefs(WuttaSet):
    """
    Form schema type for the User
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.User.roles`
    association proxy field.

    This is a subclass of :class:`WuttaSet`.  It uses a ``set`` of
    :class:`~wuttjamaican:wuttjamaican.db.model.auth.Role` ``uuid``
    values for underlying data format.
    """

    def widget_maker(self, **kwargs):
        """
        Constructs a default widget for the field.

        :returns: Instance of
           :class:`~wuttaweb.forms.widgets.RoleRefsWidget`.
        """
        session = kwargs.setdefault('session', Session())

        if 'values' not in kwargs:
            model = self.app.model
            auth = self.app.get_auth_handler()

            # avoid built-ins which cannot be assigned to users
            avoid = {
                auth.get_role_authenticated(session),
                auth.get_role_anonymous(session),
            }
            avoid = set([role.uuid for role in avoid])

            # also avoid admin unless current user is root
            if not self.request.is_root:
                avoid.add(auth.get_role_administrator(session).uuid)

            # everything else can be (un)assigned for users
            roles = session.query(model.Role)\
                           .filter(~model.Role.uuid.in_(avoid))\
                           .order_by(model.Role.name)\
                           .all()
            values = [(role.uuid.hex, role.name) for role in roles]
            kwargs['values'] = values

        return widgets.RoleRefsWidget(self.request, **kwargs)


class UserRefs(WuttaSet):
    """
    Form schema type for the Role
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.Role.users`
    association proxy field.

    This is a subclass of :class:`WuttaSet`.  It uses a ``set`` of
    :class:`~wuttjamaican:wuttjamaican.db.model.auth.User` ``uuid``
    values for underlying data format.
    """

    def widget_maker(self, **kwargs):
        """
        Constructs a default widget for the field.

        :returns: Instance of
           :class:`~wuttaweb.forms.widgets.UserRefsWidget`.
        """
        kwargs.setdefault('session', Session())
        return widgets.UserRefsWidget(self.request, **kwargs)


class Permissions(WuttaSet):
    """
    Form schema type for the Role
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.Role.permissions`
    association proxy field.

    This is a subclass of :class:`WuttaSet`.  It uses a ``set`` of
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.Permission.permission`
    values for underlying data format.

    :param permissions: Dict with all possible permissions.  Should be
       in the same format as returned by
       :meth:`~wuttaweb.views.roles.RoleView.get_available_permissions()`.
    """

    def __init__(self, request, permissions, *args, **kwargs):
        super().__init__(request, *args, **kwargs)
        self.permissions = permissions

    def widget_maker(self, **kwargs):
        """
        Constructs a default widget for the field.

        :returns: Instance of
           :class:`~wuttaweb.forms.widgets.PermissionsWidget`.
        """
        kwargs.setdefault('session', Session())
        kwargs.setdefault('permissions', self.permissions)

        if 'values' not in kwargs:
            values = []
            for gkey, group in self.permissions.items():
                for pkey, perm in group['perms'].items():
                    values.append((pkey, perm['label']))
            kwargs['values'] = values

        return widgets.PermissionsWidget(self.request, **kwargs)


class FileDownload(colander.String):
    """
    Custom schema type for a file download field.

    This field is only meant for readonly use, it does not handle file
    uploads.

    It expects the incoming ``appstruct`` to be the path to a file on
    disk (or null).

    Uses the :class:`~wuttaweb.forms.widgets.FileDownloadWidget` by
    default.

    :param request: Current :term:`request` object.

    :param url: Optional URL for hyperlink.  If not specified, file
       name/size is shown with no hyperlink.
    """

    def __init__(self, request, *args, **kwargs):
        self.url = kwargs.pop('url', None)
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def widget_maker(self, **kwargs):
        """ """
        kwargs.setdefault('url', self.url)
        return widgets.FileDownloadWidget(self.request, **kwargs)


class EmailRecipients(colander.String):
    """
    Custom schema type for :term:`email setting` recipient fields
    (``To``, ``Cc``, ``Bcc``).
    """

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return '\n'.join(parse_list(appstruct))

    def deserialize(self, node, cstruct):
        """ """
        if cstruct is colander.null:
            return colander.null

        values = [value for value in parse_list(cstruct)
                  if value]
        return ', '.join(values)

    def widget_maker(self, **kwargs):
        """
        Constructs a default widget for the field.

        :returns: Instance of
           :class:`~wuttaweb.forms.widgets.EmailRecipientsWidget`.
        """
        return widgets.EmailRecipientsWidget(**kwargs)


# nb. colanderalchemy schema overrides
sa.DateTime.__colanderalchemy_config__ = {'typ': WuttaDateTime}



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base Logic for Views
"""

import os

from pyramid import httpexceptions
from pyramid.renderers import render_to_response
from pyramid.response import FileResponse

from wuttaweb import grids


class View:
    """
    Base class for all class-based views.

    Instances of this class (or rather, a subclass) are created by
    Pyramid when processing a request.  They will have the following
    attributes:

    .. attribute:: request

       Reference to the current :term:`request` object.

    .. attribute:: app

       Reference to the :term:`app handler`.

    .. attribute:: config

       Reference to the app :term:`config object`.
    """

    def __init__(self, request, context=None):
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def forbidden(self):
        """
        Convenience method, to raise a HTTP 403 Forbidden exception::

           raise self.forbidden()
        """
        return httpexceptions.HTTPForbidden()

    def make_form(self, **kwargs):
        """
        Make and return a new :class:`~wuttaweb.forms.base.Form`
        instance, per the given ``kwargs``.

        This is the "base" factory which merely invokes the
        constructor.
        """
        web = self.app.get_web_handler()
        return web.make_form(self.request, **kwargs)

    def make_grid(self, **kwargs):
        """
        Make and return a new :class:`~wuttaweb.grids.base.Grid`
        instance, per the given ``kwargs``.

        This is the "base" factory which merely invokes the
        constructor.
        """
        web = self.app.get_web_handler()
        return web.make_grid(self.request, **kwargs)

    def make_grid_action(self, key, **kwargs):
        """
        Make and return a new :class:`~wuttaweb.grids.base.GridAction`
        instance, per the given ``key`` and ``kwargs``.

        This is the "base" factory which merely invokes the
        constructor.
        """
        return grids.GridAction(self.request, key, **kwargs)

    def notfound(self):
        """
        Convenience method, to raise a HTTP 404 Not Found exception::

           raise self.notfound()
        """
        return httpexceptions.HTTPNotFound()

    def redirect(self, url, **kwargs):
        """
        Convenience method to return a HTTP 302 response.

        Note that this technically returns an "exception" - so in
        your code, you can either return that error, or raise it::

           return self.redirect('/')
           # ..or
           raise self.redirect('/')

        Which you should do will depend on context, but raising the
        error is always "safe" since Pyramid will handle that
        correctly no matter what.
        """
        return httpexceptions.HTTPFound(location=url, **kwargs)

    def file_response(self, path, attachment=True, filename=None):
        """
        Returns a generic file response for the given path.

        :param path: Path to a file on local disk; must be accessible
           by the web app.

        :param attachment: Whether the file should come down as an
           "attachment" instead of main payload.

           The attachment behavior is the default here, and will cause
           the user to be prompted for where to save the file.

           Set ``attachment=False`` in order to cause the browser to
           render the file as if it were the page being navigated to.

        :param filename: Optional filename to use for attachment
           behavior.  This will be the "suggested filename" when user
           is prompted to save the download.  If not specified, the
           filename is derived from ``path``.

        :returns: A :class:`~pyramid:pyramid.response.FileResponse`
           object with file content.
        """
        if not os.path.exists(path):
            return self.notfound()

        response = FileResponse(path, request=self.request)
        response.content_length = os.path.getsize(path)

        if attachment:
            if not filename:
                filename = os.path.basename(path)
            response.content_disposition = f'attachment; filename="{filename}"'

        return response

    def json_response(self, context):
        """
        Returns a JSON response with the given context data.

        :param context: Context data to be rendered as JSON.

        :returns: A :term:`response` with JSON content type.
        """
        return render_to_response('json', context, request=self.request)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Event Subscribers

It is assumed that most apps will include this module somewhere during
startup.  For instance this happens within
:func:`~wuttaweb.app.main()`::

   pyramid_config.include('wuttaweb.subscribers')

This allows for certain common logic to be available for all apps.

However some custom apps may need to supplement or replace the event
hooks contained here, depending on the circumstance.
"""

import functools
import json
import logging

from pyramid import threadlocal

from wuttaweb import helpers
from wuttaweb.db import Session


log = logging.getLogger(__name__)


def new_request(event):
    """
    Event hook called when processing a new :term:`request`.

    The hook is auto-registered if this module is "included" by
    Pyramid config object.  Or you can explicitly register it::

       pyramid_config.add_subscriber('wuttaweb.subscribers.new_request',
                                     'pyramid.events.NewRequest')

    This will add to the request object:

    .. attribute:: request.wutta_config

       Reference to the app :term:`config object`.

    .. function:: request.get_referrer(default=None)

       Request method to get the "canonical" HTTP referrer value.
       This has logic to check for referrer in the request params,
       user session etc.

       :param default: Optional default URL if none is found in
          request params/session.  If no default is specified,
          the ``'home'`` route is used.

    .. attribute:: request.use_oruga

       Flag indicating whether the frontend should be displayed using
       Vue 3 + Oruga (if ``True``), or else Vue 2 + Buefy (if
       ``False``).  This flag is ``False`` by default.
    """
    request = event.request
    config = request.registry.settings['wutta_config']
    app = config.get_app()

    request.wutta_config = config

    def get_referrer(default=None):
        if request.params.get('referrer'):
            return request.params['referrer']
        if request.session.get('referrer'):
            return request.session.pop('referrer')
        referrer = getattr(request, 'referrer', None)
        if (not referrer or referrer == request.current_route_url()
            or not referrer.startswith(request.host_url)):
            referrer = default or request.route_url('home')
        return referrer

    request.get_referrer = get_referrer

    def use_oruga(request):
        spec = config.get('wuttaweb.oruga_detector.spec')
        if spec:
            func = app.load_object(spec)
            return func(request)
        return False

    request.set_property(use_oruga, reify=True)


def default_user_getter(request, db_session=None):
    """
    This is the default function used to retrieve user object from
    database.  Result of this is then assigned to :attr:`request.user`
    as part of the :func:`new_request_set_user()` hook.
    """
    uuid = request.authenticated_userid
    if uuid:
        config = request.wutta_config
        app = config.get_app()
        model = app.model
        session = db_session or Session()
        return session.get(model.User, uuid)


def new_request_set_user(
        event,
        user_getter=default_user_getter,
        db_session=None,
):
    """
    Event hook called when processing a new :term:`request`, for sake
    of setting the :attr:`request.user` and similar properties.

    The hook is auto-registered if this module is "included" by
    Pyramid config object.  Or you can explicitly register it::

       pyramid_config.add_subscriber('wuttaweb.subscribers.new_request_set_user',
                                     'pyramid.events.NewRequest')

    You may wish to "supplement" this hook by registering your own
    custom hook and then invoking this one as needed.  You can then
    pass certain params to override only parts of the logic:

    :param user_getter: Optional getter function to retrieve the user
       from database, instead of :func:`default_user_getter()`.

    :param db_session: Optional :term:`db session` to use,
       instead of :class:`wuttaweb.db.sess.Session`.

    This will add to the request object:

    .. attribute:: request.user

       Reference to the authenticated
       :class:`~wuttjamaican:wuttjamaican.db.model.auth.User` instance
       (if logged in), or ``None``.

    .. attribute:: request.is_admin

       Flag indicating whether current user is a member of the
       Administrator role.

    .. attribute:: request.is_root

       Flag indicating whether user is currently elevated to root
       privileges.  This is only possible if :attr:`request.is_admin`
       is also true.

    .. attribute:: request.user_permissions

       The ``set`` of permission names which are granted to the
       current user.

       This set is obtained by calling
       :meth:`~wuttjamaican:wuttjamaican.auth.AuthHandler.get_permissions()`.

    .. function:: request.has_perm(name)

       Shortcut to check if current user has the given permission::

          if not request.has_perm('users.edit'):
              raise self.forbidden()

    .. function:: request.has_any_perm(*names)

       Shortcut to check if current user has any of the given
       permissions::

          if request.has_any_perm('users.list', 'users.view'):
              return "can either list or view"
          else:
              raise self.forbidden()

    """
    request = event.request
    config = request.registry.settings['wutta_config']
    app = config.get_app()
    auth = app.get_auth_handler()

    # request.user
    if db_session:
        user_getter = functools.partial(user_getter, db_session=db_session)
    request.set_property(user_getter, name='user', reify=True)

    # request.is_admin
    def is_admin(request):
        return auth.user_is_admin(request.user)
    request.set_property(is_admin, reify=True)

    # request.is_root
    def is_root(request):
        if request.is_admin:
            if request.session.get('is_root', False):
                return True
        return False
    request.set_property(is_root, reify=True)

    # request.user_permissions
    def user_permissions(request):
        session = db_session or Session()
        return auth.get_permissions(session, request.user)
    request.set_property(user_permissions, reify=True)

    # request.has_perm()
    def has_perm(name):
        if request.is_root:
            return True
        if name in request.user_permissions:
            return True
        return False
    request.has_perm = has_perm

    # request.has_any_perm()
    def has_any_perm(*names):
        for name in names:
            if request.has_perm(name):
                return True
        return False
    request.has_any_perm = has_any_perm


def before_render(event):
    """
    Event hook called just before rendering a template.

    The hook is auto-registered if this module is "included" by
    Pyramid config object.  Or you can explicitly register it::

       pyramid_config.add_subscriber('wuttaweb.subscribers.before_render',
                                     'pyramid.events.BeforeRender')

    This will add some things to the template context dict.  Each of
    these may be used "directly" in a template then, e.g.:

    .. code-block:: mako

       ${app.get_title()}

    Here are the keys added to context dict by this hook:

    .. data:: 'config'

       Reference to the app :term:`config object`.

    .. data:: 'app'

       Reference to the :term:`app handler`.

    .. data:: 'web'

       Reference to the :term:`web handler`.

    .. data:: 'h'

       Reference to the helper module, :mod:`wuttaweb.helpers`.

    .. data:: 'json'

       Reference to the built-in module, :mod:`python:json`.

    .. data:: 'menus'

       Set of entries to be shown in the main menu.  This is obtained
       by calling :meth:`~wuttaweb.menus.MenuHandler.do_make_menus()`
       on the configured :class:`~wuttaweb.menus.MenuHandler`.

    .. data:: 'url'

       Reference to the request method,
       :meth:`~pyramid:pyramid.request.Request.route_url()`.
    """
    request = event.get('request') or threadlocal.get_current_request()
    config = request.wutta_config
    app = config.get_app()
    web = app.get_web_handler()

    context = event
    context['config'] = config
    context['app'] = app
    context['web'] = web
    context['h'] = helpers
    context['url'] = request.route_url
    context['json'] = json
    context['b'] = 'o' if request.use_oruga else 'b' # for buefy

    # TODO: this should be avoided somehow, for non-traditional web
    # apps, esp. "API" web apps.  (in the meantime can configure the
    # app to use NullMenuHandler which avoids most of the overhead.)
    menus = web.get_menu_handler()
    context['menus'] = menus.do_make_menus(request)


def includeme(config):
    config.add_subscriber(new_request, 'pyramid.events.NewRequest')
    config.add_subscriber(new_request_set_user, 'pyramid.events.NewRequest')
    config.add_subscriber(before_render, 'pyramid.events.BeforeRender')



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Auth Utility Logic
"""

import re

from pyramid.authentication import SessionAuthenticationHelper
from pyramid.request import RequestLocalCache
from pyramid.security import remember, forget

from wuttaweb.db import Session


def login_user(request, user):
    """
    Perform the steps necessary to "login" the given user.  This
    returns a ``headers`` dict which you should pass to the final
    redirect, like so::

       from pyramid.httpexceptions import HTTPFound

       headers = login_user(request, user)
       return HTTPFound(location='/', headers=headers)

    .. warning::

       This logic does not "authenticate" the user!  It assumes caller
       has already authenticated the user and they are safe to login.

    See also :func:`logout_user()`.
    """
    headers = remember(request, user.uuid)
    return headers


def logout_user(request):
    """
    Perform the logout action for the given request.  This returns a
    ``headers`` dict which you should pass to the final redirect, like
    so::

       from pyramid.httpexceptions import HTTPFound

       headers = logout_user(request)
       return HTTPFound(location='/', headers=headers)

    See also :func:`login_user()`.
    """
    request.session.delete()
    request.session.invalidate()
    headers = forget(request)
    return headers


class WuttaSecurityPolicy:
    """
    Pyramid :term:`security policy` for WuttaWeb.

    For more on the Pyramid details, see :doc:`pyramid:narr/security`.

    But the idea here is that you should be able to just use this,
    without thinking too hard::

       from pyramid.config import Configurator
       from wuttaweb.auth import WuttaSecurityPolicy

       pyramid_config = Configurator()
       pyramid_config.set_security_policy(WuttaSecurityPolicy())

    This security policy will then do the following:

    * use the request "web session" for auth storage (e.g. current
      ``user.uuid``)
    * check permissions as needed, by calling
      :meth:`~wuttjamaican:wuttjamaican.auth.AuthHandler.has_permission()`
      for current user

    :param db_session: Optional :term:`db session` to use, instead of
       :class:`wuttaweb.db.sess.Session`.  Probably only useful for
       tests.
    """

    def __init__(self, db_session=None):
        self.session_helper = SessionAuthenticationHelper()
        self.identity_cache = RequestLocalCache(self.load_identity)
        self.db_session = db_session or Session()

    def load_identity(self, request):
        config = request.registry.settings['wutta_config']
        app = config.get_app()
        model = app.model

        # fetch user uuid from current session
        uuid = self.session_helper.authenticated_userid(request)
        if not uuid:
            return

        # fetch user object from db
        user = self.db_session.get(model.User, uuid)
        if not user:
            return

        return user

    def identity(self, request):
        return self.identity_cache.get_or_create(request)

    def authenticated_userid(self, request):
        user = self.identity(request)
        if user is not None:
            return user.uuid

    def remember(self, request, userid, **kw):
        return self.session_helper.remember(request, userid, **kw)

    def forget(self, request, **kw):
        return self.session_helper.forget(request, **kw)

    def permits(self, request, context, permission):

        # nb. root user can do anything
        if getattr(request, 'is_root', False):
            return True

        config = request.registry.settings['wutta_config']
        app = config.get_app()
        auth = app.get_auth_handler()
        user = self.identity(request)
        return auth.has_permission(self.db_session, user, permission)


def add_permission_group(pyramid_config, groupkey, label=None, overwrite=True):
    """
    Pyramid directive to add a "permission group" to the app's
    awareness.

    The app must be made aware of all permissions, so they are exposed
    when editing a
    :class:`~wuttjamaican:wuttjamaican.db.model.auth.Role`.  The logic
    for discovering permissions is in
    :meth:`~wuttaweb.views.roles.RoleView.get_available_permissions()`.

    This is usually called from within a master view's
    :meth:`~wuttaweb.views.master.MasterView.defaults()` to establish
    the permission group which applies to the view model.

    A simple example of usage::

       pyramid_config.add_permission_group('widgets', label="Widgets")

    :param groupkey: Unique key for the permission group.  In the
       context of a master view, this will be the same as
       :attr:`~wuttaweb.views.master.MasterView.permission_prefix`.

    :param label: Optional label for the permission group.  If not
       specified, it is derived from ``groupkey``.

    :param overwrite: If the permission group was already established,
       this flag controls whether the group's label should be
       overwritten (with ``label``).

    See also :func:`add_permission()`.
    """
    config = pyramid_config.get_settings()['wutta_config']
    app = config.get_app()
    def action():
        perms = pyramid_config.get_settings().get('wutta_permissions', {})
        if overwrite or groupkey not in perms:
            group = perms.setdefault(groupkey, {'key': groupkey})
            group['label'] = label or app.make_title(groupkey)
        pyramid_config.add_settings({'wutta_permissions': perms})
    pyramid_config.action(None, action)


def add_permission(pyramid_config, groupkey, key, label=None):
    """
    Pyramid directive to add a single "permission" to the app's
    awareness.

    The app must be made aware of all permissions, so they are exposed
    when editing a
    :class:`~wuttjamaican:wuttjamaican.db.model.auth.Role`.  The logic
    for discovering permissions is in
    :meth:`~wuttaweb.views.roles.RoleView.get_available_permissions()`.

    This is usually called from within a master view's
    :meth:`~wuttaweb.views.master.MasterView.defaults()` to establish
    "known" permissions based on master view feature flags
    (:attr:`~wuttaweb.views.master.MasterView.viewable`,
    :attr:`~wuttaweb.views.master.MasterView.editable`, etc.).

    A simple example of usage::

       pyramid_config.add_permission('widgets', 'widgets.polish',
                                     label="Polish all the widgets")

    :param groupkey: Unique key for the permission group.  In the
       context of a master view, this will be the same as
       :attr:`~wuttaweb.views.master.MasterView.permission_prefix`.

    :param key: Unique key for the permission.  This should be the
       "complete" permission name which includes the permission
       prefix.

    :param label: Optional label for the permission.  If not
       specified, it is derived from ``key``.

    See also :func:`add_permission_group()`.
    """
    def action():
        config = pyramid_config.get_settings()['wutta_config']
        app = config.get_app()
        perms = pyramid_config.get_settings().get('wutta_permissions', {})
        group = perms.setdefault(groupkey, {'key': groupkey})
        group.setdefault('label', app.make_title(groupkey))
        perm = group.setdefault('perms', {}).setdefault(key, {'key': key})
        perm['label'] = label or app.make_title(key)
        pyramid_config.add_settings({'wutta_permissions': perms})
    pyramid_config.action(None, action)
